# This file is part of cloud-init. See LICENSE file for license information.

"""Tests for handling of userdata within cloud init."""

import gzip
import logging
import os
from email import encoders
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from io import BytesIO
from pathlib import Path
from unittest import mock

import pytest
import requests
import responses

from cloudinit import handlers
from cloudinit import helpers as c_helpers
from cloudinit import safeyaml, stages
from cloudinit import user_data as ud
from cloudinit import util
from cloudinit.config.modules import Modules
from cloudinit.settings import PER_INSTANCE
from tests.unittests import helpers
from tests.unittests.util import FakeDataSource

MPATH = "cloudinit.stages"


def count_messages(root):
    am = 0
    for m in root.walk():
        if ud.is_skippable(m):
            continue
        am += 1
    return am


def gzip_text(text):
    contents = BytesIO()
    f = gzip.GzipFile(fileobj=contents, mode="wb")
    f.write(util.encode_text(text))
    f.flush()
    f.close()
    return contents.getvalue()


@pytest.fixture(scope="function")
def init_tmp(request, tmpdir):
    ci = stages.Init()
    cloud_dir = tmpdir.join("cloud")
    cloud_dir.mkdir()
    run_dir = tmpdir.join("run")
    run_dir.mkdir()
    ci._cfg = {
        "system_info": {
            "default_user": {"name": "ubuntu"},
            "distro": "ubuntu",
            "paths": {
                "cloud_dir": cloud_dir.strpath,
                "run_dir": run_dir.strpath,
            },
        }
    }
    run_dir.join("instance-data-sensitive.json").write("{}")
    return ci


class TestConsumeUserData:
    def test_simple_jsonp(self, init_tmp):
        user_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "qux" },
     { "op": "add", "path": "/bar", "value": "qux2" }
]
"""
        init_tmp.datasource = FakeDataSource(user_blob)
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset"):
            init_tmp.consume_data()
        cc_contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        cc = util.load_yaml(cc_contents)
        assert len(cc) == 2
        assert cc["baz"] == "qux"
        assert cc["bar"] == "qux2"

    @pytest.mark.usefixtures("fake_filesystem")
    def test_simple_jsonp_vendor_and_vendor2_and_user(self):
        # test that user-data wins over vendor
        user_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "qux" },
     { "op": "add", "path": "/bar", "value": "qux2" },
     { "op": "add", "path": "/foobar", "value": "qux3" }
]
"""
        vendor_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "quxA" },
     { "op": "add", "path": "/bar", "value": "quxB" },
     { "op": "add", "path": "/foo", "value": "quxC" },
     { "op": "add", "path": "/corge", "value": "quxEE" }
]
"""
        vendor2_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/corge", "value": "quxD" },
     { "op": "add", "path": "/grault", "value": "quxFF" },
     { "op": "add", "path": "/foobar", "value": "quxGG" }
]
"""
        initer = stages.Init()
        initer.datasource = FakeDataSource(
            user_blob, vendordata=vendor_blob, vendordata2=vendor2_blob
        )
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        with mock.patch(
            "cloudinit.util.read_conf_from_cmdline", return_value={}
        ):
            initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mods = Modules(initer)
        (_which_ran, _failures) = mods.run_section("cloud_init_modules")
        cfg = mods.cfg
        assert "vendor_data" in cfg
        assert "vendor_data2" in cfg
        # Confirm that vendordata2 overrides vendordata, and that
        #  userdata overrides both
        assert cfg["baz"] == "qux"
        assert cfg["bar"] == "qux2"
        assert cfg["foobar"] == "qux3"
        assert cfg["foo"] == "quxC"
        assert cfg["corge"] == "quxD"
        assert cfg["grault"] == "quxFF"

    @pytest.mark.usefixtures("fake_filesystem")
    def test_simple_jsonp_no_vendor_consumed(self):
        # make sure that vendor data is not consumed
        user_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "qux" },
     { "op": "add", "path": "/bar", "value": "qux2" },
     { "op": "add", "path": "/vendor_data", "value": {"enabled": "false"}}
]
"""
        vendor_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "quxA" },
     { "op": "add", "path": "/bar", "value": "quxB" },
     { "op": "add", "path": "/foo", "value": "quxC" }
]
"""
        initer = stages.Init()
        initer.datasource = FakeDataSource(user_blob, vendordata=vendor_blob)
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mods = Modules(initer)
        (_which_ran, _failures) = mods.run_section("cloud_init_modules")
        cfg = mods.cfg
        assert cfg["baz"] == "qux"
        assert cfg["bar"] == "qux2"
        assert "foo" not in cfg

    def test_mixed_cloud_config(self, init_tmp):
        blob_cc = """
#cloud-config
a: b
c: d
"""
        message_cc = MIMEBase("text", "cloud-config")
        message_cc.set_payload(blob_cc)

        blob_jp = """
#cloud-config-jsonp
[
     { "op": "replace", "path": "/a", "value": "c" },
     { "op": "remove", "path": "/c" }
]
"""

        message_jp = MIMEBase("text", "cloud-config-jsonp")
        message_jp.set_payload(blob_jp)

        message = MIMEMultipart()
        message.attach(message_cc)
        message.attach(message_jp)

        init_tmp.datasource = FakeDataSource(str(message))
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset"):
            init_tmp.consume_data()
        cc_contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        cc = util.load_yaml(cc_contents)
        assert len(cc) == 1
        assert cc["a"] == "c"

    def test_cloud_config_as_x_shell_script(self, init_tmp):
        blob_cc = """
#cloud-config
a: b
c: d
"""
        message_cc = MIMEBase("text", "x-shellscript")
        message_cc.set_payload(blob_cc)

        blob_jp = """
#cloud-config-jsonp
[
     { "op": "replace", "path": "/a", "value": "c" },
     { "op": "remove", "path": "/c" }
]
"""

        message_jp = MIMEBase("text", "cloud-config-jsonp")
        message_jp.set_payload(blob_jp)

        message = MIMEMultipart()
        message.attach(message_cc)
        message.attach(message_jp)

        init_tmp.datasource = FakeDataSource(str(message))
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset"):
            init_tmp.consume_data()
        cc_contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        cc = util.load_yaml(cc_contents)
        assert len(cc) == 1
        assert cc["a"] == "c"

    @pytest.mark.usefixtures("fake_filesystem")
    def test_vendor_user_yaml_cloud_config(self):
        vendor_blob = """
#cloud-config
a: b
name: vendor
run:
 - x
 - y
"""

        user_blob = """
#cloud-config
a: c
vendor_data:
  enabled: true
  prefix: /bin/true
name: user
run:
 - z
"""
        initer = stages.Init()
        initer.datasource = FakeDataSource(user_blob, vendordata=vendor_blob)
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mods = Modules(initer)
        (_which_ran, _failures) = mods.run_section("cloud_init_modules")
        cfg = mods.cfg
        assert "vendor_data" in cfg
        assert cfg["a"] == "c"
        assert cfg["name"] == "user"
        assert "x" not in cfg["run"]
        assert "y" not in cfg["run"]
        assert "z" in cfg["run"]

    @pytest.mark.usefixtures("fake_filesystem")
    def test_vendordata_script(self):
        vendor_blob = """
#!/bin/bash
echo "test"
"""
        vendor2_blob = """
#!/bin/bash
echo "dynamic test"
"""

        user_blob = """
#cloud-config
vendor_data:
  enabled: true
  prefix: /bin/true
"""
        initer = stages.Init()
        initer.datasource = FakeDataSource(
            user_blob, vendordata=vendor_blob, vendordata2=vendor2_blob
        )
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mods = Modules(initer)
        (_which_ran, _failures) = mods.run_section("cloud_init_modules")
        vendor_script = initer.paths.get_ipath_cur("vendor_scripts")
        vendor_script_fns = "%s/part-001" % vendor_script
        assert os.path.exists(vendor_script_fns) is True

    def test_merging_cloud_config(self, tmpdir):
        blob = """
#cloud-config
a: b
e: f
run:
 - b
 - c
"""
        message1 = MIMEBase("text", "cloud-config")
        message1.set_payload(blob)

        blob2 = """
#cloud-config
a: e
e: g
run:
 - stuff
 - morestuff
"""
        message2 = MIMEBase("text", "cloud-config")
        message2[
            "X-Merge-Type"
        ] = "dict(recurse_array,recurse_str)+list(append)+str(append)"
        message2.set_payload(blob2)

        blob3 = """
#cloud-config
e:
 - 1
 - 2
 - 3
p: 1
"""
        message3 = MIMEBase("text", "cloud-config")
        message3.set_payload(blob3)

        messages = [message1, message2, message3]

        paths = c_helpers.Paths(
            {"cloud_dir": tmpdir, "run_dir": tmpdir}, ds=FakeDataSource("")
        )
        cloud_cfg = handlers.cloud_config.CloudConfigPartHandler(paths)

        cloud_cfg.handle_part(
            None, handlers.CONTENT_START, None, None, None, None
        )
        for i, m in enumerate(messages):
            headers = dict(m)
            fn = "part-%s" % (i + 1)
            payload = m.get_payload(decode=True)
            cloud_cfg.handle_part(
                None, headers["Content-Type"], fn, payload, None, headers
            )
        cloud_cfg.handle_part(
            None, handlers.CONTENT_END, None, None, None, None
        )
        contents = util.load_text_file(paths.get_ipath("cloud_config"))
        contents = util.load_yaml(contents)
        assert contents["run"], ["b", "c", "stuff", "morestuff"]
        assert contents["a"] == "be"
        assert contents["e"] == [1, 2, 3]
        assert contents["p"] == 1

    def test_unhandled_type_warning(self, init_tmp, caplog):
        """Raw text without magic is ignored but shows warning."""
        data = "arbitrary text\n"
        init_tmp.datasource = FakeDataSource(data)

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
            assert (
                "Unhandled non-multipart (text/x-not-multipart) userdata:"
                in caplog.text
            )
        mockobj.assert_called_once_with(
            init_tmp.paths.get_ipath("cloud_config"), "", 0o600
        )

    def test_mime_gzip_compressed(self, init_tmp):
        """Tests that individual message gzip encoding works."""

        def gzip_part(text):
            return MIMEApplication(gzip_text(text), "gzip")

        base_content1 = """
#cloud-config
a: 2
"""

        base_content2 = """
#cloud-config
b: 3
c: 4
"""

        message = MIMEMultipart("test")
        message.attach(gzip_part(base_content1))
        message.attach(gzip_part(base_content2))
        init_tmp.datasource = FakeDataSource(str(message))
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset"):
            init_tmp.consume_data()
        contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        contents = util.load_yaml(contents)
        assert isinstance(contents, dict) is True
        assert len(contents) == 3
        assert contents["a"] == 2
        assert contents["b"] == 3
        assert contents["c"] == 4

    def test_mime_text_plain(self, init_tmp, caplog):
        """Mime message of type text/plain is ignored but shows warning."""
        message = MIMEBase("text", "plain")
        message.set_payload("Just text")
        init_tmp.datasource = FakeDataSource(message.as_string().encode())

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
            assert "Unhandled unknown content-type (text/plain)" in caplog.text
        mockobj.assert_called_once_with(
            init_tmp.paths.get_ipath("cloud_config"), "", 0o600
        )

    # Since features are intended to be overridden downstream, mock them
    # all here so new feature flags don't require a new change to this
    # unit test.
    @mock.patch.multiple(
        "cloudinit.features",
        ERROR_ON_USER_DATA_FAILURE=True,
        ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES=True,
        EXPIRE_APPLIES_TO_HASHED_USERS=False,
        NETPLAN_CONFIG_ROOT_READ_ONLY=True,
        NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH=False,
        APT_DEB822_SOURCE_LIST_FILE=True,
    )
    def test_shellscript(self, init_tmp, tmpdir, caplog):
        """Raw text starting #!/bin/sh is treated as script."""
        script = "#!/bin/sh\necho hello\n"
        init_tmp.datasource = FakeDataSource(script)

        outpath = os.path.join(
            init_tmp.paths.get_ipath_cur("scripts"), "part-001"
        )

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
                assert caplog.records == []  # No warnings

        mockobj.assert_has_calls(
            [
                mock.call(outpath, script, 0o700),
                mock.call(init_tmp.paths.get_ipath("cloud_config"), "", 0o600),
            ]
        )
        expected = {
            "features": {
                "ERROR_ON_USER_DATA_FAILURE": True,
                "ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES": True,
                "EXPIRE_APPLIES_TO_HASHED_USERS": False,
                "NETPLAN_CONFIG_ROOT_READ_ONLY": True,
                "NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH": False,
                "APT_DEB822_SOURCE_LIST_FILE": True,
            },
            "system_info": {
                "default_user": {"name": "ubuntu"},
                "distro": "ubuntu",
                "paths": {
                    "cloud_dir": tmpdir.join("cloud").strpath,
                    "run_dir": tmpdir.join("run").strpath,
                },
            },
        }

        loaded_json = util.load_json(
            util.load_text_file(
                init_tmp.paths.get_runpath("instance_data_sensitive")
            )
        )
        assert expected == loaded_json

        expected["_doc"] = stages.COMBINED_CLOUD_CONFIG_DOC
        assert expected == util.load_json(
            util.load_text_file(
                init_tmp.paths.get_runpath("combined_cloud_config")
            )
        )

    def test_mime_text_x_shellscript(self, init_tmp, caplog):
        """Mime message of type text/x-shellscript is treated as script."""
        script = "#!/bin/sh\necho hello\n"
        message = MIMEBase("text", "x-shellscript")
        message.set_payload(script)
        init_tmp.datasource = FakeDataSource(message.as_string())

        outpath = os.path.join(
            init_tmp.paths.get_ipath_cur("scripts"), "part-001"
        )

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
                assert caplog.records == []  # No warnings

        mockobj.assert_has_calls(
            [
                mock.call(outpath, script, 0o700),
                mock.call(init_tmp.paths.get_ipath("cloud_config"), "", 0o600),
            ]
        )

    def test_mime_text_plain_shell(self, init_tmp, caplog):
        """Mime type text/plain starting #!/bin/sh is treated as script."""
        script = "#!/bin/sh\necho hello\n"
        message = MIMEBase("text", "plain")
        message.set_payload(script)
        init_tmp.datasource = FakeDataSource(message.as_string())

        outpath = os.path.join(
            init_tmp.paths.get_ipath_cur("scripts"), "part-001"
        )

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
                assert caplog.records == []  # No warnings

        mockobj.assert_has_calls(
            [
                mock.call(outpath, script, 0o700),
                mock.call(init_tmp.paths.get_ipath("cloud_config"), "", 0o600),
            ]
        )

    def test_mime_application_octet_stream(self, init_tmp, caplog):
        """Mime type application/octet-stream is ignored but shows warning."""
        message = MIMEBase("application", "octet-stream")
        message.set_payload(b"\xbf\xe6\xb2\xc3\xd3\xba\x13\xa4\xd8\xa1\xcc")
        encoders.encode_base64(message)
        init_tmp.datasource = FakeDataSource(message.as_string().encode())

        with mock.patch("cloudinit.util.write_file") as mockobj:
            with caplog.at_level(logging.WARNING):
                init_tmp.fetch()
                with mock.patch.object(init_tmp, "_reset"):
                    init_tmp.consume_data()
                    assert (
                        "Unhandled unknown content-type"
                        " (application/octet-stream)" in caplog.text
                    )
        mockobj.assert_called_once_with(
            init_tmp.paths.get_ipath("cloud_config"), "", 0o600
        )

    def test_cloud_config_archive(self, init_tmp):
        non_decodable = b"\x11\xc9\xb4gTH\xee\x12"
        data = [
            {"content": "#cloud-config\npassword: gocubs\n"},
            {"content": "#cloud-config\nlocale: chicago\n"},
            {"content": non_decodable},
        ]
        message = b"#cloud-config-archive\n" + safeyaml.dumps(data).encode()

        init_tmp.datasource = FakeDataSource(message)

        fs = {}

        def fsstore(filename, content, mode=0o0644, omode="wb"):
            fs[filename] = content

        # consuming the user-data provided should write 'cloud_config' file
        # which will have our yaml in it.
        with mock.patch("cloudinit.util.write_file") as mockobj:
            mockobj.side_effect = fsstore
            init_tmp.fetch()
            with mock.patch.object(init_tmp, "_reset"):
                init_tmp.consume_data()

        cfg = util.load_yaml(fs[init_tmp.paths.get_ipath("cloud_config")])
        assert cfg.get("password") == "gocubs"
        assert cfg.get("locale") == "chicago"

    @pytest.mark.usefixtures("fake_filesystem")
    @mock.patch("cloudinit.util.read_conf_with_confd")
    def test_dont_allow_user_data(self, mock_cfg):
        mock_cfg.return_value = {"allow_userdata": False}

        # test that user-data is ignored but vendor-data is kept
        user_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "qux" },
     { "op": "add", "path": "/bar", "value": "qux2" }
]
"""
        vendor_blob = """
#cloud-config-jsonp
[
     { "op": "add", "path": "/baz", "value": "quxA" },
     { "op": "add", "path": "/bar", "value": "quxB" },
     { "op": "add", "path": "/foo", "value": "quxC" }
]
"""
        init = stages.Init()
        init.datasource = FakeDataSource(user_blob, vendordata=vendor_blob)
        init.read_cfg()
        init.initialize()
        init.fetch()
        init.instancify()
        init.update()
        init.cloudify().run(
            "consume_data",
            init.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mods = Modules(init)
        (_which_ran, _failures) = mods.run_section("cloud_init_modules")
        cfg = mods.cfg
        assert "vendor_data" in cfg
        assert cfg["baz"] == "quxA"
        assert cfg["bar"] == "quxB"
        assert cfg["foo"] == "quxC"


class TestConsumeUserDataHttp:
    @responses.activate
    @mock.patch("cloudinit.url_helper.time.sleep")
    def test_include(self, mock_sleep, init_tmp):
        """Test #include."""
        included_url = "http://hostname/path"
        included_data = "#cloud-config\nincluded: true\n"
        responses.add(responses.GET, included_url, included_data)

        init_tmp.datasource = FakeDataSource("#include\nhttp://hostname/path")
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset") as _reset:
            init_tmp.consume_data()
            assert _reset.call_count == 1
        cc_contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        cc = util.load_yaml(cc_contents)
        assert cc.get("included") is True

    @responses.activate
    @mock.patch("cloudinit.url_helper.time.sleep")
    def test_include_bad_url(self, mock_sleep, init_tmp):
        """Test #include with a bad URL."""
        bad_url = "http://bad/forbidden"
        bad_data = "#cloud-config\nbad: true\n"
        responses.add(responses.GET, bad_url, bad_data, status=403)

        included_url = "http://hostname/path"
        included_data = "#cloud-config\nincluded: true\n"
        responses.add(responses.GET, included_url, included_data)

        init_tmp.datasource = FakeDataSource(
            "#include\nhttp://bad/forbidden\nhttp://hostname/path"
        )
        init_tmp.fetch()
        with pytest.raises(Exception, match="403"):
            with mock.patch.object(init_tmp, "_reset") as _reset:
                init_tmp.consume_data()
                assert _reset.call_count == 1

        with pytest.raises(FileNotFoundError):
            util.load_text_file(init_tmp.paths.get_ipath("cloud_config"))

    @responses.activate
    @mock.patch("cloudinit.url_helper.time.sleep")
    @mock.patch("cloudinit.util.is_container")
    @mock.patch(
        "cloudinit.user_data.features.ERROR_ON_USER_DATA_FAILURE", False
    )
    def test_include_bad_url_no_fail(
        self, is_container, mock_sleep, tmpdir, init_tmp, caplog
    ):
        """Test #include with a bad URL and failure disabled"""
        is_container.return_value = True
        bad_url = "http://bad/forbidden"
        responses.add(
            responses.GET,
            bad_url,
            body=requests.HTTPError(
                f"403 Client Error: Forbidden for url: {bad_url}"
            ),
            status=403,
        )

        included_url = "http://hostname/path"
        included_data = "#cloud-config\nincluded: true\n"
        responses.add(responses.GET, included_url, included_data)

        init_tmp.datasource = FakeDataSource(
            "#include\nhttp://bad/forbidden\nhttp://hostname/path"
        )
        init_tmp.fetch()
        with mock.patch.object(init_tmp, "_reset") as _reset:
            init_tmp.consume_data()
            assert _reset.call_count == 1

        assert (
            "403 Client Error: Forbidden for url: %s" % bad_url in caplog.text
        )

        cc_contents = util.load_text_file(
            init_tmp.paths.get_ipath("cloud_config")
        )
        cc = util.load_yaml(cc_contents)
        assert cc.get("bad") is None
        assert cc.get("included") is True


class TestUDProcess(helpers.ResourceUsingTestCase):
    def test_bytes_in_userdata(self):
        msg = b"#cloud-config\napt_update: True\n"
        ud_proc = ud.UserDataProcessor(self.getCloudPaths())
        message = ud_proc.process(msg)
        self.assertTrue(count_messages(message) == 1)

    def test_string_in_userdata(self):
        msg = "#cloud-config\napt_update: True\n"

        ud_proc = ud.UserDataProcessor(self.getCloudPaths())
        message = ud_proc.process(msg)
        self.assertTrue(count_messages(message) == 1)

    def test_compressed_in_userdata(self):
        msg = gzip_text("#cloud-config\napt_update: True\n")

        ud_proc = ud.UserDataProcessor(self.getCloudPaths())
        message = ud_proc.process(msg)
        self.assertTrue(count_messages(message) == 1)


class TestConvertString(helpers.TestCase):
    def test_handles_binary_non_utf8_decodable(self):
        """Printable unicode (not utf8-decodable) is safely converted."""
        blob = b"#!/bin/bash\necho \xc3\x84\n"
        msg = ud.convert_string(blob)
        self.assertEqual(blob, msg.get_payload(decode=True))

    def test_handles_binary_utf8_decodable(self):
        blob = b"\x32\x32"
        msg = ud.convert_string(blob)
        self.assertEqual(blob, msg.get_payload(decode=True))

    def test_handle_headers(self):
        text = "hi mom"
        msg = ud.convert_string(text)
        self.assertEqual(text, msg.get_payload(decode=False))

    def test_handle_mime_parts(self):
        """Mime parts are properly returned as a mime message."""
        message = MIMEBase("text", "plain")
        message.set_payload("Just text")
        msg = ud.convert_string(str(message))
        self.assertEqual("Just text", msg.get_payload(decode=False))


class TestFetchBaseConfig:
    @pytest.fixture(autouse=True)
    def mocks(self, mocker):
        mocker.patch(f"{MPATH}.util.read_conf_from_cmdline")
        mocker.patch(f"{MPATH}.read_runtime_config")

    def test_only_builtin_gets_builtin(self, mocker):
        mocker.patch(f"{MPATH}.read_runtime_config", return_value={})
        mocker.patch(f"{MPATH}.util.read_conf_with_confd")
        config = stages.fetch_base_config()
        assert util.get_builtin_cfg() == config

    def test_conf_d_overrides_defaults(self, mocker):
        builtin = util.get_builtin_cfg()
        test_key = sorted(builtin)[0]
        test_value = "test"

        mocker.patch(
            f"{MPATH}.util.read_conf_with_confd",
            return_value={test_key: test_value},
        )
        mocker.patch(f"{MPATH}.read_runtime_config", return_value={})
        config = stages.fetch_base_config()
        assert config.get(test_key) == test_value
        builtin[test_key] = test_value
        assert config == builtin

    def test_confd_with_template(self, mocker, tmp_path: Path):
        instance_data_path = tmp_path / "test_confd_with_template.json"
        instance_data_path.write_text('{"template_var": "template_value"}')
        cfg_path = tmp_path / "test_conf_with_template.cfg"
        cfg_path.write_text('## template:jinja\n{"key": "{{template_var}}"}')

        mocker.patch("cloudinit.stages.CLOUD_CONFIG", cfg_path)
        mocker.patch(f"{MPATH}.util.get_builtin_cfg", return_value={})
        config = stages.fetch_base_config(
            instance_data_file=instance_data_path
        )
        assert config == {"key": "template_value"}

    def test_cmdline_overrides_defaults(self, mocker):
        builtin = util.get_builtin_cfg()
        test_key = sorted(builtin)[0]
        test_value = "test"
        cmdline = {test_key: test_value}

        mocker.patch(f"{MPATH}.util.read_conf_with_confd")
        mocker.patch(
            f"{MPATH}.util.read_conf_from_cmdline",
            return_value=cmdline,
        )
        mocker.patch(f"{MPATH}.read_runtime_config")
        config = stages.fetch_base_config()
        assert config.get(test_key) == test_value
        builtin[test_key] = test_value
        assert config == builtin

    def test_cmdline_overrides_confd_runtime_and_defaults(self, mocker):
        builtin = {"key1": "value0", "key3": "other2"}
        conf_d = {"key1": "value1", "key2": "other1"}
        cmdline = {"key3": "other3", "key2": "other2"}
        runtime = {"key3": "runtime3"}

        mocker.patch(f"{MPATH}.util.read_conf_with_confd", return_value=conf_d)
        mocker.patch(f"{MPATH}.util.get_builtin_cfg", return_value=builtin)
        mocker.patch(f"{MPATH}.read_runtime_config", return_value=runtime)
        mocker.patch(
            f"{MPATH}.util.read_conf_from_cmdline",
            return_value=cmdline,
        )

        config = stages.fetch_base_config()
        assert config == {"key1": "value1", "key2": "other2", "key3": "other3"}

    def test_order_precedence_is_builtin_system_runtime_cmdline(self, mocker):
        builtin = {"key1": "builtin0", "key3": "builtin3"}
        conf_d = {"key1": "confd1", "key2": "confd2", "keyconfd1": "kconfd1"}
        runtime = {"key1": "runtime1", "key2": "runtime2"}
        cmdline = {"key1": "cmdline1"}

        mocker.patch(f"{MPATH}.util.read_conf_with_confd", return_value=conf_d)
        mocker.patch(f"{MPATH}.util.get_builtin_cfg", return_value=builtin)
        mocker.patch(
            f"{MPATH}.util.read_conf_from_cmdline",
            return_value=cmdline,
        )
        mocker.patch(f"{MPATH}.read_runtime_config", return_value=runtime)

        config = stages.fetch_base_config()

        assert config == {
            "key1": "cmdline1",
            "key2": "runtime2",
            "key3": "builtin3",
            "keyconfd1": "kconfd1",
        }



"""Integration test for gh-626.

Ensure if wakeonlan is specified in the network config that it is rendered
in the /etc/network/interfaces or netplan config.
"""
import pytest
import yaml

from tests.integration_tests import random_mac_address
from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.integration_settings import PLATFORM

MAC_ADDRESS = random_mac_address()
NETWORK_CONFIG = """\
version: 2
ethernets:
  eth0:
    dhcp4: true
    wakeonlan: true
    match:
      {match_clause}
"""

# On netplan 0.106 and above, netplan emits mac address matches as
# PermanentMACAdddress on networkd systems. This match clause will
# not match LXD's veth devices which results in no IPv4 address
# being allocated via DHCP. As a result, we match container by NIC name.
# LP: #2022947 represents this netplan behavior
NETWORK_CONFIG_CONTAINER = NETWORK_CONFIG.format(match_clause="name: eth0")

# Match LXD VM by MAC just to assert plumbing is working by MAC
NETWORK_CONFIG_VM = NETWORK_CONFIG.format(
    match_clause=f"macaddress: {MAC_ADDRESS}"
)

EXPECTED_ENI_END = """\
iface eth0 inet dhcp
    ethernet-wol g"""


@pytest.mark.skipif(
    PLATFORM not in ["lxd_container", "lxd_vm"],
    reason="Test requires custom networking provided by LXD",
)
def test_wakeonlan(session_cloud: IntegrationCloud):
    if PLATFORM == "lxd_vm":
        config_dict = {
            "user.network-config": NETWORK_CONFIG_VM,
            "volatile.eth0.hwaddr": MAC_ADDRESS,
        }
    else:
        config_dict = {"user.network-config": NETWORK_CONFIG_CONTAINER}
    with session_cloud.launch(
        launch_kwargs={
            "config_dict": config_dict,
        }
    ) as client:
        netplan_cfg = client.execute("cat /etc/netplan/50-cloud-init.yaml")
        netplan_yaml = yaml.safe_load(netplan_cfg)
        assert "wakeonlan" in netplan_yaml["network"]["ethernets"]["eth0"]
        assert (
            netplan_yaml["network"]["ethernets"]["eth0"]["wakeonlan"] is True
        )



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import os
import re
import shutil
import tempfile

from cloudinit import util
from cloudinit.config import cc_apt_configure
from tests.unittests.helpers import TestCase


class TestAptProxyConfig(TestCase):
    def setUp(self):
        super(TestAptProxyConfig, self).setUp()
        self.tmp = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmp)
        self.pfile = os.path.join(self.tmp, "proxy.cfg")
        self.cfile = os.path.join(self.tmp, "config.cfg")

    def _search_apt_config(self, contents, ptype, value):
        return re.search(
            r"acquire::%s::proxy\s+[\"']%s[\"'];\n" % (ptype, value),
            contents,
            flags=re.IGNORECASE,
        )

    def test_apt_proxy_written(self):
        cfg = {"proxy": "myproxy"}
        cc_apt_configure.apply_apt_config(cfg, self.pfile, self.cfile)

        self.assertTrue(os.path.isfile(self.pfile))
        self.assertFalse(os.path.isfile(self.cfile))

        contents = util.load_text_file(self.pfile)
        self.assertTrue(self._search_apt_config(contents, "http", "myproxy"))

    def test_apt_http_proxy_written(self):
        cfg = {"http_proxy": "myproxy"}
        cc_apt_configure.apply_apt_config(cfg, self.pfile, self.cfile)

        self.assertTrue(os.path.isfile(self.pfile))
        self.assertFalse(os.path.isfile(self.cfile))

        contents = util.load_text_file(self.pfile)
        self.assertTrue(self._search_apt_config(contents, "http", "myproxy"))

    def test_apt_all_proxy_written(self):
        cfg = {
            "http_proxy": "myproxy_http_proxy",
            "https_proxy": "myproxy_https_proxy",
            "ftp_proxy": "myproxy_ftp_proxy",
        }

        values = {
            "http": cfg["http_proxy"],
            "https": cfg["https_proxy"],
            "ftp": cfg["ftp_proxy"],
        }

        cc_apt_configure.apply_apt_config(cfg, self.pfile, self.cfile)

        self.assertTrue(os.path.isfile(self.pfile))
        self.assertFalse(os.path.isfile(self.cfile))

        contents = util.load_text_file(self.pfile)

        for ptype, pval in values.items():
            self.assertTrue(self._search_apt_config(contents, ptype, pval))

    def test_proxy_deleted(self):
        util.write_file(self.cfile, "content doesnt matter")
        cc_apt_configure.apply_apt_config({}, self.pfile, self.cfile)
        self.assertFalse(os.path.isfile(self.pfile))
        self.assertFalse(os.path.isfile(self.cfile))

    def test_proxy_replaced(self):
        util.write_file(self.cfile, "content doesnt matter")
        cc_apt_configure.apply_apt_config(
            {"proxy": "foo"}, self.pfile, self.cfile
        )
        self.assertTrue(os.path.isfile(self.pfile))
        contents = util.load_text_file(self.pfile)
        self.assertTrue(self._search_apt_config(contents, "http", "foo"))

    def test_config_written(self):
        payload = "this is my apt config"
        cfg = {"conf": payload}

        cc_apt_configure.apply_apt_config(cfg, self.pfile, self.cfile)

        self.assertTrue(os.path.isfile(self.cfile))
        self.assertFalse(os.path.isfile(self.pfile))

        self.assertEqual(util.load_text_file(self.cfile), payload)

    def test_config_replaced(self):
        util.write_file(self.pfile, "content doesnt matter")
        cc_apt_configure.apply_apt_config(
            {"conf": "foo"}, self.pfile, self.cfile
        )
        self.assertTrue(os.path.isfile(self.cfile))
        self.assertEqual(util.load_text_file(self.cfile), "foo")

    def test_config_deleted(self):
        # if no 'conf' is provided, delete any previously written file
        util.write_file(self.pfile, "content doesnt matter")
        cc_apt_configure.apply_apt_config({}, self.pfile, self.cfile)
        self.assertFalse(os.path.isfile(self.pfile))
        self.assertFalse(os.path.isfile(self.cfile))


class TestConversion(TestCase):
    def test_convert_with_apt_mirror_as_empty_string(self):
        # an empty apt_mirror is the same as no apt_mirror
        empty_m_found = cc_apt_configure.convert_to_v3_apt_format(
            {"apt_mirror": ""}
        )
        default_found = cc_apt_configure.convert_to_v3_apt_format({})
        self.assertEqual(default_found, empty_m_found)

    def test_convert_with_apt_mirror(self):
        mirror = "http://my.mirror/ubuntu"
        f = cc_apt_configure.convert_to_v3_apt_format({"apt_mirror": mirror})
        self.assertIn(mirror, set(m["uri"] for m in f["apt"]["primary"]))

    def test_no_old_content(self):
        mirror = "http://my.mirror/ubuntu"
        mydata = {"apt": {"primary": {"arches": ["default"], "uri": mirror}}}
        expected = copy.deepcopy(mydata)
        self.assertEqual(
            expected, cc_apt_configure.convert_to_v3_apt_format(mydata)
        )



# Copyright (C) 2014 Vaidas Jablonskis
#
# Author: Vaidas Jablonskis <jablonskis@gmail.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import datetime
import json
import re
from base64 import b64decode, b64encode
from unittest import mock
from urllib.parse import urlparse

import responses

from cloudinit import distros, helpers, settings
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.sources import DataSourceGCE
from tests.unittests import helpers as test_helpers

M_PATH = "cloudinit.sources.DataSourceGCE."

GCE_META = {
    "instance/id": "123",
    "instance/zone": "foo/bar",
    "instance/hostname": "server.project-foo.local",
}

GCE_META_PARTIAL = {
    "instance/id": "1234",
    "instance/hostname": "server.project-bar.local",
    "instance/zone": "bar/baz",
}

GCE_META_ENCODING = {
    "instance/id": "12345",
    "instance/hostname": "server.project-baz.local",
    "instance/zone": "baz/bang",
    "instance/attributes": {
        "user-data": b64encode(b"#!/bin/echo baz\n").decode("utf-8"),
        "user-data-encoding": "base64",
    },
}

GCE_USER_DATA_TEXT = {
    "instance/id": "12345",
    "instance/hostname": "server.project-baz.local",
    "instance/zone": "baz/bang",
    "instance/attributes": {
        "user-data": "#!/bin/sh\necho hi mom\ntouch /run/up-now\n",
    },
}

HEADERS = {"Metadata-Flavor": "Google"}
MD_URL_RE = re.compile(
    r"http://metadata.google.internal/computeMetadata/v1/.*"
)
GUEST_ATTRIBUTES_URL = (
    "http://metadata.google.internal/computeMetadata/"
    "v1/instance/guest-attributes/hostkeys/"
)


class TestDataSourceGCE(test_helpers.ResponsesTestCase):
    with_logs = True

    def _make_distro(self, dtype, def_user=None):
        cfg = dict(settings.CFG_BUILTIN)
        cfg["system_info"]["distro"] = dtype
        paths = helpers.Paths(cfg["system_info"]["paths"])
        distro_cls = distros.fetch(dtype)
        if def_user:
            cfg["system_info"]["default_user"] = def_user.copy()
        distro = distro_cls(dtype, cfg["system_info"], paths)
        return distro

    def setUp(self):
        tmp = self.tmp_dir()
        self.ds = DataSourceGCE.DataSourceGCE(
            settings.CFG_BUILTIN, None, helpers.Paths({"run_dir": tmp})
        )

        ppatch = self.m_platform_reports_gce = mock.patch(
            M_PATH + "platform_reports_gce"
        )
        self.m_platform_reports_gce = ppatch.start()
        self.m_platform_reports_gce.return_value = True
        self.addCleanup(ppatch.stop)

        pppatch = self.m_is_resolvable_url = mock.patch(
            M_PATH + "util.is_resolvable_url", return_value=True
        )
        self.m_is_resolvable_url = pppatch.start()
        self.addCleanup(pppatch.stop)

        self.add_patch("time.sleep", "m_sleep")  # just to speed up tests
        super(TestDataSourceGCE, self).setUp()

    def _set_mock_metadata(self, gce_meta=None, *, check_headers=None):
        if gce_meta is None:
            gce_meta = GCE_META

        def _request_callback(request):
            url_path = urlparse(request.url).path
            if url_path.startswith("/computeMetadata/v1/"):
                path = url_path.split("/computeMetadata/v1/")[1:][0]
                recursive = path.endswith("/")
                path = path.rstrip("/")
            else:
                path = None
            if path in gce_meta:
                response = gce_meta.get(path)
                if recursive:
                    response = json.dumps(response)
                if check_headers is not None:
                    for k in check_headers.keys():
                        self.assertEqual(check_headers[k], request.headers[k])
                return (200, request.headers, response)
            else:
                return (404, request.headers, "")

        self.responses.add_callback(
            responses.GET,
            MD_URL_RE,
            callback=_request_callback,
        )

    def test_connection(self):
        self._set_mock_metadata(check_headers=HEADERS)
        success = self.ds.get_data()
        self.assertTrue(success)

    def test_metadata(self):
        # UnicodeDecodeError if set to ds.userdata instead of userdata_raw
        meta = GCE_META.copy()
        meta["instance/attributes/user-data"] = b"/bin/echo \xff\n"

        self._set_mock_metadata()
        self.ds.get_data()

        shostname = GCE_META.get("instance/hostname").split(".")[0]
        self.assertEqual(shostname, self.ds.get_hostname().hostname)

        self.assertEqual(
            GCE_META.get("instance/id"), self.ds.get_instance_id()
        )

        self.assertEqual(
            GCE_META.get("instance/attributes/user-data"),
            self.ds.get_userdata_raw(),
        )

    def test_metadata_partial(self):
        """test partial metadata (missing user-data in particular)"""
        self._set_mock_metadata(GCE_META_PARTIAL)
        self.ds.get_data()

        self.assertEqual(
            GCE_META_PARTIAL.get("instance/id"), self.ds.get_instance_id()
        )

        shostname = GCE_META_PARTIAL.get("instance/hostname").split(".")[0]
        self.assertEqual(shostname, self.ds.get_hostname().hostname)

    def test_userdata_no_encoding(self):
        """check that user-data is read."""
        self._set_mock_metadata(GCE_USER_DATA_TEXT)
        self.ds.get_data()
        self.assertEqual(
            GCE_USER_DATA_TEXT["instance/attributes"]["user-data"].encode(),
            self.ds.get_userdata_raw(),
        )

    def test_metadata_encoding(self):
        """user-data is base64 encoded if user-data-encoding is 'base64'."""
        self._set_mock_metadata(GCE_META_ENCODING)
        self.ds.get_data()

        instance_data = GCE_META_ENCODING.get("instance/attributes")
        decoded = b64decode(instance_data.get("user-data"))
        self.assertEqual(decoded, self.ds.get_userdata_raw())

    def test_missing_required_keys_return_false(self):
        for required_key in [
            "instance/id",
            "instance/zone",
            "instance/hostname",
        ]:
            meta = GCE_META_PARTIAL.copy()
            del meta[required_key]
            self._set_mock_metadata(meta)
            self.assertEqual(False, self.ds.get_data())
            self.responses.reset()

    def test_no_ssh_keys_metadata(self):
        self._set_mock_metadata()
        self.ds.get_data()
        self.assertEqual([], self.ds.get_public_ssh_keys())

    def test_cloudinit_ssh_keys(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(0)),
                    "user:{0}".format(invalid_key.format(0)),
                ]
            ),
            "ssh-keys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(1)),
                    "user:{0}".format(invalid_key.format(1)),
                ]
            ),
        }
        instance_attributes = {
            "ssh-keys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(2)),
                    "user:{0}".format(invalid_key.format(2)),
                ]
            ),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(key) for key in range(3)]
        self.assertEqual(set(expected), set(self.ds.get_public_ssh_keys()))

    @mock.patch(M_PATH + "ug_util")
    def test_default_user_ssh_keys(self, mock_ug_util):
        mock_ug_util.normalize_users_groups.return_value = None, None
        mock_ug_util.extract_default.return_value = "ubuntu", None
        ubuntu_ds = DataSourceGCE.DataSourceGCE(
            settings.CFG_BUILTIN,
            self._make_distro("ubuntu"),
            helpers.Paths({"run_dir": self.tmp_dir()}),
        )

        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(0)),
                    "user:{0}".format(invalid_key.format(0)),
                ]
            ),
            "ssh-keys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(1)),
                    "user:{0}".format(invalid_key.format(1)),
                ]
            ),
        }
        instance_attributes = {
            "ssh-keys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(2)),
                    "user:{0}".format(invalid_key.format(2)),
                ]
            ),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        ubuntu_ds.get_data()

        expected = [valid_key.format(key) for key in range(3)]
        self.assertEqual(set(expected), set(ubuntu_ds.get_public_ssh_keys()))

    def test_instance_ssh_keys_override(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "cloudinit:{0}".format(invalid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(invalid_key.format(1)),
        }
        instance_attributes = {
            "sshKeys": "cloudinit:{0}".format(valid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(valid_key.format(1)),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(key) for key in range(2)]
        self.assertEqual(set(expected), set(self.ds.get_public_ssh_keys()))

    def test_block_project_ssh_keys_override(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "cloudinit:{0}".format(invalid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(invalid_key.format(1)),
        }
        instance_attributes = {
            "ssh-keys": "cloudinit:{0}".format(valid_key.format(0)),
            "block-project-ssh-keys": "True",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(0)]
        self.assertEqual(set(expected), set(self.ds.get_public_ssh_keys()))

    def test_only_last_part_of_zone_used_for_availability_zone(self):
        self._set_mock_metadata()
        r = self.ds.get_data()
        self.assertEqual(True, r)
        self.assertEqual("bar", self.ds.availability_zone)

    @mock.patch("cloudinit.sources.DataSourceGCE.GoogleMetadataFetcher")
    def test_get_data_returns_false_if_not_on_gce(self, m_fetcher):
        self.m_platform_reports_gce.return_value = False
        ret = self.ds.get_data()
        self.assertEqual(False, ret)
        m_fetcher.assert_not_called()

    def test_has_expired(self):
        def _get_timestamp(days):
            format_str = "%Y-%m-%dT%H:%M:%S+0000"
            today = datetime.datetime.now()
            timestamp = today + datetime.timedelta(days=days)
            return timestamp.strftime(format_str)

        past = _get_timestamp(-1)
        future = _get_timestamp(1)
        ssh_keys = {
            None: False,
            "": False,
            "Invalid": False,
            "user:ssh-rsa key user@domain.com": False,
            'user:ssh-rsa key google {"expireOn":"%s"}' % past: False,
            "user:ssh-rsa key google-ssh": False,
            "user:ssh-rsa key google-ssh {invalid:json}": False,
            'user:ssh-rsa key google-ssh {"userName":"user"}': False,
            'user:ssh-rsa key google-ssh {"expireOn":"invalid"}': False,
            'user:xyz key google-ssh {"expireOn":"%s"}' % future: False,
            'user:xyz key google-ssh {"expireOn":"%s"}' % past: True,
        }

        for key, expired in ssh_keys.items():
            self.assertEqual(DataSourceGCE._has_expired(key), expired)

    def test_parse_public_keys_non_ascii(self):
        public_key_data = [
            "cloudinit:rsa ssh-ke%s invalid" % chr(165),
            "use%sname:rsa ssh-key" % chr(174),
            "cloudinit:test 1",
            "default:test 2",
            "user:test 3",
        ]
        expected = ["test 1", "test 2"]
        found = DataSourceGCE._parse_public_keys(
            public_key_data, default_user="default"
        )
        self.assertEqual(sorted(found), sorted(expected))

    @mock.patch("cloudinit.url_helper.readurl")
    def test_publish_host_keys(self, m_readurl):
        hostkeys = [("ssh-rsa", "asdfasdf"), ("ssh-ed25519", "qwerqwer")]
        readurl_expected_calls = [
            mock.call(
                check_status=False,
                data=b"asdfasdf",
                headers=HEADERS,
                request_method="PUT",
                url="%s%s" % (GUEST_ATTRIBUTES_URL, "ssh-rsa"),
            ),
            mock.call(
                check_status=False,
                data=b"qwerqwer",
                headers=HEADERS,
                request_method="PUT",
                url="%s%s" % (GUEST_ATTRIBUTES_URL, "ssh-ed25519"),
            ),
        ]
        self.ds.publish_host_keys(hostkeys)
        m_readurl.assert_has_calls(readurl_expected_calls, any_order=True)

    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    @mock.patch(M_PATH + "net.find_candidate_nics", return_value=["ens4"])
    def test_local_datasource_uses_ephemeral_dhcp(
        self, _m_find_candidate_nics, m_dhcp
    ):
        self._set_mock_metadata()
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = self.tmp_dir
        ds = DataSourceGCE.DataSourceGCELocal(
            sys_cfg={}, distro=distro, paths=None
        )
        ds._get_data()
        assert m_dhcp.call_count == 1

    @mock.patch(M_PATH + "util.log_time")
    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    @mock.patch(M_PATH + "net.find_candidate_nics")
    def test_local_datasource_tries_on_multi_nic(
        self, m_find_candidate_nics, m_dhcp, m_read_md
    ):
        self._set_mock_metadata()
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = self.tmp_dir
        ds = DataSourceGCE.DataSourceGCELocal(
            sys_cfg={}, distro=distro, paths=None
        )
        m_find_candidate_nics.return_value = [
            "ens0p4",
            "ens0p5",
            "ens0p6",
            "ens4",
        ]
        m_dhcp.return_value.__enter__.side_effect = (
            None,
            NoDHCPLeaseError,
            None,
            None,
        )
        m_read_md.side_effect = (
            {"success": False},
            Exception("whoopsie, not this one"),
            {
                "success": True,
                "platform_reports_gce": True,
                "reason": "reason",
                "user-data": "ud",
                "meta-data": "md",
            },
        )
        assert ds._get_data() is True
        assert m_dhcp.call_args_list == [
            mock.call(distro, iface=DataSourceGCE.DEFAULT_PRIMARY_INTERFACE),
            mock.call(distro, iface="ens0p4"),
            mock.call(distro, iface="ens0p5"),
            mock.call(distro, iface="ens0p6"),
        ]
        assert ds.distro.fallback_interface == "ens0p6"
        assert ds.metadata == "md"
        assert ds.userdata_raw == "ud"

        expected_logs = (
            "Looking for the primary NIC in:"
            " ['ens4', 'ens0p4', 'ens0p5', 'ens0p6']",
            "Error fetching IMD with candidate NIC ens0p5:"
            " whoopsie, not this one",
        )
        for msg in expected_logs:
            self.assertIn(
                msg,
                self.logs.getvalue(),
            )

    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    def test_datasource_doesnt_use_ephemeral_dhcp(self, m_dhcp):
        self._set_mock_metadata()
        ds = DataSourceGCE.DataSourceGCE(sys_cfg={}, distro=None, paths=None)
        ds._get_data()
        assert m_dhcp.call_count == 0



#!/usr/bin/env python3

# Copyright (C) 2017 Canonical Ltd.
#
# This file is part of cloud-init. See LICENSE file for license information.

"""Define 'collect-logs' utility and handler to include in cloud-init cmd."""

import argparse
import os
import shutil
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import NamedTuple

from cloudinit.cmd.devel import read_cfg_paths
from cloudinit.helpers import Paths
from cloudinit.stages import Init
from cloudinit.subp import ProcessExecutionError, subp
from cloudinit.temp_utils import tempdir
from cloudinit.util import (
    chdir,
    copy,
    ensure_dir,
    get_config_logfiles,
    write_file,
)

CLOUDINIT_RUN_DIR = "/run/cloud-init"


class ApportFile(NamedTuple):
    path: str
    label: str


INSTALLER_APPORT_SENSITIVE_FILES = [
    ApportFile(
        "/var/log/installer/autoinstall-user-data", "AutoInstallUserData"
    ),
    ApportFile("/autoinstall.yaml", "AutoInstallYAML"),
    ApportFile("/etc/cloud/cloud.cfg.d/99-installer.cfg", "InstallerCloudCfg"),
]

INSTALLER_APPORT_FILES = [
    ApportFile("/var/log/installer/ubuntu_desktop_installer.log", "UdiLog"),
    ApportFile(
        "/var/log/installer/subiquity-server-debug.log", "SubiquityServerDebug"
    ),
    ApportFile(
        "/var/log/installer/subiquity-client-debug.log", "SubiquityClientDebug"
    ),
    ApportFile("/var/log/installer/curtin-install.log", "CurtinLog"),
    # Legacy single curtin config < 22.1
    ApportFile(
        "/var/log/installer/subiquity-curtin-install.conf",
        "CurtinInstallConfig",
    ),
    ApportFile(
        "/var/log/installer/curtin-install/subiquity-initial.conf",
        "CurtinConfigInitial",
    ),
    ApportFile(
        "/var/log/installer/curtin-install/subiquity-curthooks.conf",
        "CurtinConfigCurtHooks",
    ),
    ApportFile(
        "/var/log/installer/curtin-install/subiquity-extract.conf",
        "CurtinConfigExtract",
    ),
    ApportFile(
        "/var/log/installer/curtin-install/subiquity-partitioning.conf",
        "CurtinConfigPartitioning",
    ),
    # Legacy curtin < 22.1 curtin error tar path
    ApportFile("/var/log/installer/curtin-error-logs.tar", "CurtinError"),
    ApportFile("/var/log/installer/curtin-errors.tar", "CurtinError"),
    ApportFile("/var/log/installer/block/probe-data.json", "ProbeData"),
]


def _get_user_data_file() -> str:
    paths = read_cfg_paths()
    return paths.get_ipath_cur("userdata_raw")


def _get_cloud_data_path() -> str:
    paths = read_cfg_paths()
    return paths.get_cpath("data")


def get_parser(parser=None):
    """Build or extend and arg parser for collect-logs utility.

    @param parser: Optional existing ArgumentParser instance representing the
        collect-logs subcommand which will be extended to support the args of
        this utility.

    @returns: ArgumentParser with proper argument configuration.
    """
    if not parser:
        parser = argparse.ArgumentParser(
            prog="collect-logs",
            description="Collect and tar all cloud-init debug info",
        )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        dest="verbosity",
        help="Be more verbose.",
    )
    parser.add_argument(
        "--tarfile",
        "-t",
        default="cloud-init.tar.gz",
        help=(
            "The tarfile to create containing all collected logs."
            " Default: cloud-init.tar.gz"
        ),
    )
    user_data_file = _get_user_data_file()
    parser.add_argument(
        "--include-userdata",
        "-u",
        default=False,
        action="store_true",
        dest="userdata",
        help=(
            "Optionally include user-data from {0} which could contain"
            " sensitive information.".format(user_data_file)
        ),
    )
    return parser


def _copytree_rundir_ignore_files(curdir, files):
    """Return a list of files to ignore for /run/cloud-init directory"""
    ignored_files = [
        "hook-hotplug-cmd",  # named pipe for hotplug
    ]
    if os.getuid() != 0:
        # Ignore root-permissioned files
        ignored_files.append(Paths({}).lookups["instance_data_sensitive"])
    return ignored_files


def _write_command_output_to_file(cmd, filename, msg, verbosity):
    """Helper which runs a command and writes output or error to filename."""
    ensure_dir(os.path.dirname(filename))
    try:
        output = subp(cmd).stdout
    except ProcessExecutionError as e:
        write_file(filename, str(e))
        _debug("collecting %s failed.\n" % msg, 1, verbosity)
    else:
        write_file(filename, output)
        _debug("collected %s\n" % msg, 1, verbosity)
        return output


def _stream_command_output_to_file(cmd, filename, msg, verbosity):
    """Helper which runs a command and writes output or error to filename."""
    ensure_dir(os.path.dirname(filename))
    try:
        with open(filename, "w") as f:
            subprocess.call(cmd, stdout=f, stderr=f)
    except OSError as e:
        write_file(filename, str(e))
        _debug("collecting %s failed.\n" % msg, 1, verbosity)
    else:
        _debug("collected %s\n" % msg, 1, verbosity)


def _debug(msg, level, verbosity):
    if level <= verbosity:
        sys.stderr.write(msg)


def _collect_file(path, out_dir, verbosity):
    if os.path.isfile(path):
        copy(path, out_dir)
        _debug("collected file: %s\n" % path, 1, verbosity)
    else:
        _debug("file %s did not exist\n" % path, 2, verbosity)


def collect_installer_logs(log_dir, include_userdata, verbosity):
    """Obtain subiquity logs and config files."""
    for src_file in INSTALLER_APPORT_FILES:
        destination_dir = Path(log_dir + src_file.path).parent
        if not destination_dir.exists():
            ensure_dir(str(destination_dir))
        _collect_file(src_file.path, str(destination_dir), verbosity)
    if include_userdata:
        for src_file in INSTALLER_APPORT_SENSITIVE_FILES:
            destination_dir = Path(log_dir + src_file.path).parent
            if not destination_dir.exists():
                ensure_dir(str(destination_dir))
            _collect_file(src_file.path, str(destination_dir), verbosity)


def collect_logs(tarfile, include_userdata: bool, verbosity=0):
    """Collect all cloud-init logs and tar them up into the provided tarfile.

    @param tarfile: The path of the tar-gzipped file to create.
    @param include_userdata: Boolean, true means include user-data.
    """
    if include_userdata and os.getuid() != 0:
        sys.stderr.write(
            "To include userdata, root user is required."
            " Try sudo cloud-init collect-logs\n"
        )
        return 1

    init = Init(ds_deps=[])
    tarfile = os.path.abspath(tarfile)
    log_dir = datetime.utcnow().date().strftime("cloud-init-logs-%Y-%m-%d")
    with tempdir(dir="/tmp") as tmp_dir:
        log_dir = os.path.join(tmp_dir, log_dir)
        version = _write_command_output_to_file(
            cmd=["cloud-init", "--version"],
            filename=os.path.join(log_dir, "version"),
            msg="cloud-init --version",
            verbosity=verbosity,
        )
        dpkg_ver = _write_command_output_to_file(
            cmd=["dpkg-query", "--show", "-f=${Version}\n", "cloud-init"],
            filename=os.path.join(log_dir, "dpkg-version"),
            msg="dpkg version",
            verbosity=verbosity,
        )
        if not version:
            version = dpkg_ver if dpkg_ver else "not-available"
        print("version: ", version)
        _debug("collected cloud-init version: %s\n" % version, 1, verbosity)
        _stream_command_output_to_file(
            cmd=["dmesg"],
            filename=os.path.join(log_dir, "dmesg.txt"),
            msg="dmesg output",
            verbosity=verbosity,
        )
        _stream_command_output_to_file(
            cmd=["journalctl", "--boot=0", "-o", "short-precise"],
            filename=os.path.join(log_dir, "journal.txt"),
            msg="systemd journal of current boot",
            verbosity=verbosity,
        )

        init.read_cfg()
        for log in get_config_logfiles(init.cfg):
            _collect_file(log, log_dir, verbosity)
        if include_userdata:
            user_data_file = _get_user_data_file()
            _collect_file(user_data_file, log_dir, verbosity)
        collect_installer_logs(log_dir, include_userdata, verbosity)

        run_dir = os.path.join(log_dir, "run")
        ensure_dir(run_dir)
        if os.path.exists(CLOUDINIT_RUN_DIR):
            try:
                shutil.copytree(
                    CLOUDINIT_RUN_DIR,
                    os.path.join(run_dir, "cloud-init"),
                    ignore=_copytree_rundir_ignore_files,
                )
            except shutil.Error as e:
                sys.stderr.write("Failed collecting file(s) due to error:\n")
                sys.stderr.write(str(e) + "\n")
            _debug("collected dir %s\n" % CLOUDINIT_RUN_DIR, 1, verbosity)
        else:
            _debug(
                "directory '%s' did not exist\n" % CLOUDINIT_RUN_DIR,
                1,
                verbosity,
            )
        if os.path.exists(os.path.join(CLOUDINIT_RUN_DIR, "disabled")):
            # Fallback to grab previous cloud/data
            cloud_data_dir = Path(_get_cloud_data_path())
            if cloud_data_dir.exists():
                shutil.copytree(
                    str(cloud_data_dir),
                    Path(log_dir + str(cloud_data_dir)),
                )
        with chdir(tmp_dir):
            subp(["tar", "czvf", tarfile, log_dir.replace(tmp_dir + "/", "")])
    sys.stderr.write("Wrote %s\n" % tarfile)
    return 0


def handle_collect_logs_args(name, args):
    """Handle calls to 'cloud-init collect-logs' as a subcommand."""
    return collect_logs(args.tarfile, args.userdata, args.verbosity)


def main():
    """Tool to collect and tar all cloud-init related logs."""
    parser = get_parser()
    return handle_collect_logs_args("collect-logs", parser.parse_args())


if __name__ == "__main__":
    sys.exit(main())



# This file is part of cloud-init. See LICENSE file for license information.
"""
This file is for testing the feature flag functionality itself,
NOT for testing any individual feature flag
"""
from unittest import mock

from cloudinit import features


class TestGetFeatures:
    def test_feature_without_override(self):
        # Since features are intended to be overridden downstream, mock them
        # all here so new feature flags don't require a new change to this
        # unit test.
        with mock.patch.multiple(
            "cloudinit.features",
            ERROR_ON_USER_DATA_FAILURE=True,
            ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES=True,
            EXPIRE_APPLIES_TO_HASHED_USERS=False,
            NETPLAN_CONFIG_ROOT_READ_ONLY=True,
            NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH=False,
            APT_DEB822_SOURCE_LIST_FILE=True,
        ):
            assert {
                "ERROR_ON_USER_DATA_FAILURE": True,
                "ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES": True,
                "EXPIRE_APPLIES_TO_HASHED_USERS": False,
                "NETPLAN_CONFIG_ROOT_READ_ONLY": True,
                "NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH": False,
                "APT_DEB822_SOURCE_LIST_FILE": True,
            } == features.get_features()



# This file is part of cloud-init. See LICENSE file for license information.

"""Tests for testing reporting and event handling."""

import json

import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.util import ASSETS_DIR, verify_clean_log

URL = "http://127.0.0.1:55555"

USER_DATA = f"""\
#cloud-config
reporting:
  webserver:
    type: webhook
    endpoint: "{URL}"
    timeout: 1
    retries: 1

"""


@pytest.mark.user_data(USER_DATA)
def test_webhook_reporting(client: IntegrationInstance):
    """Test when using webhook reporting that we get expected events.

    This test setups a simple echo server that prints out POST data out to
    a file. Ensure that that file contains all of the expected events.
    """
    client.push_file(ASSETS_DIR / "echo_server.py", "/var/tmp/echo_server.py")
    client.push_file(
        ASSETS_DIR / "echo_server.service",
        "/etc/systemd/system/echo_server.service",
    )
    client.execute("cloud-init clean --logs")
    client.execute("systemctl start echo_server.service")
    # Run through our standard process here. This remove any uncertainty
    # around messages transmitting during pre-network boot.
    client.execute(
        "cloud-init init --local; "
        "cloud-init init; "
        "cloud-init modules --mode=config; "
        "cloud-init modules --mode=final; "
        "cloud-init status --wait"
    )
    verify_clean_log(client.read_from_file("/var/log/cloud-init.log"))

    server_output = client.read_from_file(
        "/var/tmp/echo_server_output"
    ).splitlines()
    events = [json.loads(line) for line in server_output]

    # Only time this should be less is if we remove modules
    assert len(events) > 52, events

    # Assert our first and last expected messages exist
    ds_events = [
        e for e in events if e["name"] == "init-network/activate-datasource"
    ]
    assert len(ds_events) == 2  # 1 for start, 1 for stop

    final_events = [e for e in events if e["name"] == "modules-final"]
    assert final_events  # 1 for stop and ignore LP: #1992711 for now



# Copyright (C) 2012 Canonical Ltd.
#
# Author: Ben Howard <ben.howard@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"Users and Groups: Configure users and groups"

import logging
from textwrap import dedent

from cloudinit.cloud import Cloud

# Ensure this is aliased to a name not 'distros'
# since the module attribute 'distros'
# is a list of distros that are supported, not a sub-module
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.distros import ug_util
from cloudinit.settings import PER_INSTANCE

MODULE_DESCRIPTION = """\
This module configures users and groups. For more detailed information on user
options, see the :ref:`Including users and groups<yaml_examples>` config
example.

Groups to add to the system can be specified under the ``groups`` key as
a string of comma-separated groups to create, or a list. Each item in
the list should either contain a string of a single group to create,
or a dictionary with the group name as the key and string of a single user as
a member of that group or a list of users who should be members of the group.

.. note::
   Groups are added before users, so any users in a group list must
   already exist on the system.

Users to add can be specified as a string or list under the ``users`` key.
Each entry in the list should either be a string or a dictionary. If a string
is specified, that string can be comma-separated usernames to create or the
reserved string ``default`` which represents the primary admin user used to
access the system. The ``default`` user varies per distribution and is
generally configured in ``/etc/cloud/cloud.cfg`` by the ``default_user`` key.

Each ``users`` dictionary item must contain either a ``name`` or ``snapuser``
key, otherwise it will be ignored. Omission of ``default`` as the first item
in the ``users`` list skips creation the default user. If no ``users`` key is
provided the default behavior is to create the default user via this config::

 users:
 - default

.. note::
    Specifying a hash of a user's password with ``passwd`` is a security risk
    if the cloud-config can be intercepted. SSH authentication is preferred.

.. note::
    If specifying a doas rule for a user, ensure that the syntax for the rule
    is valid, as the only checking performed by cloud-init is to ensure that
    the user referenced in the rule is the correct user.

.. note::
    If specifying a sudo rule for a user, ensure that the syntax for the rule
    is valid, as it is not checked by cloud-init.

.. note::
    Most of these configuration options will not be honored if the user
    already exists. The following options are the exceptions; they are applied
    to already-existing users: ``plain_text_passwd``, ``doas``,
    ``hashed_passwd``, ``lock_passwd``, ``sudo``, ``ssh_authorized_keys``,
    ``ssh_redirect_user``.

The ``user`` key can be used to override the ``default_user`` configuration
defined in ``/etc/cloud/cloud.cfg``. The ``user`` value should be a dictionary
which supports the same config keys as the ``users`` dictionary items.
"""

meta: MetaSchema = {
    "id": "cc_users_groups",
    "name": "Users and Groups",
    "title": "Configure users and groups",
    "description": MODULE_DESCRIPTION,
    "distros": ["all"],
    "examples": [
        dedent(
            """\
        # Add the ``default_user`` from /etc/cloud/cloud.cfg.
        # This is also the default behavior of cloud-init when no `users` key
        # is provided.
        users:
        - default
        """
        ),
        dedent(
            """\
        # Add the 'admingroup' with members 'root' and 'sys' and an empty
        # group cloud-users.
        groups:
        - admingroup: [root,sys]
        - cloud-users
        """
        ),
        dedent(
            """\
        # Skip creation of the <default> user and only create newsuper.
        # Password-based login is rejected, but the github user TheRealFalcon
        # and the launchpad user falcojr can SSH as newsuper. The default
        # shell for newsuper is bash instead of system default.
        users:
        - name: newsuper
          gecos: Big Stuff
          groups: users, admin
          sudo: ALL=(ALL) NOPASSWD:ALL
          shell: /bin/bash
          lock_passwd: true
          ssh_import_id:
            - lp:falcojr
            - gh:TheRealFalcon
        """
        ),
        dedent(
            """\
        # Skip creation of the <default> user and only create newsuper.
        # Password-based login is rejected, but the github user TheRealFalcon
        # and the launchpad user falcojr can SSH as newsuper. doas/opendoas
        # is configured to permit this user to run commands as other users
        # (without being prompted for a password) except not as root.
        users:
        - name: newsuper
          gecos: Big Stuff
          groups: users, admin
          doas:
            - permit nopass newsuper
            - deny newsuper as root
          lock_passwd: true
          ssh_import_id:
            - lp:falcojr
            - gh:TheRealFalcon
        """
        ),
        dedent(
            """\
        # On a system with SELinux enabled, add youruser and set the
        # SELinux user to 'staff_u'. When omitted on SELinux, the system will
        # select the configured default SELinux user.
        users:
        - default
        - name: youruser
          selinux_user: staff_u
        """
        ),
        dedent(
            """\
        # To redirect a legacy username to the <default> user for a
        # distribution, ssh_redirect_user will accept an SSH connection and
        # emit a message telling the client to ssh as the <default> user.
        # SSH clients will get the message:
        users:
        - default
        - name: nosshlogins
          ssh_redirect_user: true
        """
        ),
        dedent(
            """\
        # Override any ``default_user`` config in /etc/cloud/cloud.cfg with
        # supplemental config options.
        # This config will make the default user to mynewdefault and change
        # the user to not have sudo rights.
        ssh_import_id: [chad.smith]
        user:
          name: mynewdefault
          sudo: null
        """
        ),
        dedent(
            """\
        # Avoid creating any ``default_user``.
        users: []
        """
        ),
    ],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": [],
}

__doc__ = get_meta_doc(meta)

LOG = logging.getLogger(__name__)

# NO_HOME and NEED_HOME are mutually exclusive options
NO_HOME = ("no_create_home", "system")
NEED_HOME = ("ssh_authorized_keys", "ssh_import_id", "ssh_redirect_user")


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    (users, groups) = ug_util.normalize_users_groups(cfg, cloud.distro)
    (default_user, _user_config) = ug_util.extract_default(users)
    cloud_keys = cloud.get_public_ssh_keys() or []

    for (name, members) in groups.items():
        cloud.distro.create_group(name, members)

    for (user, config) in users.items():

        no_home = [key for key in NO_HOME if config.get(key)]
        need_home = [key for key in NEED_HOME if config.get(key)]
        if no_home and need_home:
            raise ValueError(
                f"Not creating user {user}. Key(s) {', '.join(need_home)}"
                f" cannot be provided with {', '.join(no_home)}"
            )

        ssh_redirect_user = config.pop("ssh_redirect_user", False)
        if ssh_redirect_user:
            if "ssh_authorized_keys" in config or "ssh_import_id" in config:
                raise ValueError(
                    "Not creating user %s. ssh_redirect_user cannot be"
                    " provided with ssh_import_id or ssh_authorized_keys"
                    % user
                )
            if ssh_redirect_user not in (True, "default"):
                raise ValueError(
                    "Not creating user %s. Invalid value of"
                    " ssh_redirect_user: %s. Expected values: true, default"
                    " or false." % (user, ssh_redirect_user)
                )
            if default_user is None:
                LOG.warning(
                    "Ignoring ssh_redirect_user: %s for %s."
                    " No default_user defined."
                    " Perhaps missing cloud configuration users: "
                    " [default, ..].",
                    ssh_redirect_user,
                    user,
                )
            else:
                config["ssh_redirect_user"] = default_user
                config["cloud_public_ssh_keys"] = cloud_keys

        cloud.distro.create_user(user, **config)



import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.releases import CURRENT_RELEASE

USER_DATA = """\
#cloud-config
runcmd:
 - echo "hi" >> /var/tmp/hi
"""


@pytest.mark.user_data(USER_DATA)
def test_frequency_override(client: IntegrationInstance):
    # Some pre-checks
    assert (
        "running config-scripts_user with frequency once-per-instance"
        in client.read_from_file("/var/log/cloud-init.log")
    )
    assert client.read_from_file("/var/tmp/hi").strip().count("hi") == 1
    if CURRENT_RELEASE.os == "ubuntu":
        if CURRENT_RELEASE.series in ("focal", "jammy", "lunar", "mantic"):
            # Stable series will block on snapd.seeded.service and create a
            # semaphore file
            assert client.execute("test -f /var/lib/cloud/snap-seeded.once").ok
        else:
            # Newer series will not block on snapd.seeded.service nor create a
            # semaphore file
            assert not client.execute(
                "test -f /var/lib/cloud/snap-seeded.once"
            ).ok

    # Change frequency of scripts_user to always
    config = client.read_from_file("/etc/cloud/cloud.cfg")
    new_config = config.replace("- scripts_user", "- [ scripts_user, always ]")
    client.write_to_file("/etc/cloud/cloud.cfg", new_config)

    client.restart()

    # Ensure the script was run again
    assert (
        "running config-scripts_user with frequency always"
        in client.read_from_file("/var/log/cloud-init.log")
    )
    assert client.read_from_file("/var/tmp/hi").strip().count("hi") == 2



"""NoCloud datasource integration tests."""
from textwrap import dedent

import pytest
from pycloudlib.lxd.instance import LXDInstance

from cloudinit.subp import subp
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM

VENDOR_DATA = """\
#cloud-config
runcmd:
  - touch /var/tmp/seeded_vendordata_test_file
"""


LXD_METADATA_NOCLOUD_SEED = """\
  /var/lib/cloud/seed/nocloud-net/meta-data:
    when:
    - create
    - copy
    create_only: false
    template: emptycfg.tpl
    properties:
      default: |
        #cloud-config
        {}
  /var/lib/cloud/seed/nocloud-net/user-data:
    when:
    - create
    - copy
    create_only: false
    template: emptycfg.tpl
    properties:
      default: |
        #cloud-config
        {}
"""


def setup_nocloud(instance: LXDInstance):
    # On Jammy and above, LXD no longer uses NoCloud, so we need to set
    # it up manually
    lxd_image_metadata = subp(
        ["lxc", "config", "metadata", "show", instance.name]
    )
    if "/var/lib/cloud/seed/nocloud-net" in lxd_image_metadata.stdout:
        return
    subp(
        ["lxc", "config", "template", "create", instance.name, "emptycfg.tpl"],
    )
    subp(
        ["lxc", "config", "template", "edit", instance.name, "emptycfg.tpl"],
        data="#cloud-config\n{}\n",
    )
    subp(
        ["lxc", "config", "metadata", "edit", instance.name],
        data=f"{lxd_image_metadata.stdout}{LXD_METADATA_NOCLOUD_SEED}",
    )


@pytest.mark.lxd_setup.with_args(setup_nocloud)
@pytest.mark.lxd_use_exec
@pytest.mark.skipif(
    PLATFORM != "lxd_container",
    reason="Requires NoCloud with custom setup",
)
def test_nocloud_seedfrom_vendordata(client: IntegrationInstance):
    """Integration test for #570.

    Test that we can add optional vendor-data to the seedfrom file in a
    NoCloud environment
    """
    seed_dir = "/var/tmp/test_seed_dir"
    result = client.execute(
        "mkdir {seed_dir} && "
        "touch {seed_dir}/user-data && "
        "touch {seed_dir}/meta-data && "
        "echo 'seedfrom: {seed_dir}/' > "
        "/var/lib/cloud/seed/nocloud-net/meta-data".format(seed_dir=seed_dir)
    )
    assert result.return_code == 0

    client.write_to_file(
        "{}/vendor-data".format(seed_dir),
        VENDOR_DATA,
    )
    client.execute("cloud-init clean --logs")
    client.restart()
    assert client.execute("cloud-init status").ok
    assert "seeded_vendordata_test_file" in client.execute("ls /var/tmp")


SMBIOS_USERDATA = """\
#cloud-config
runcmd:
  - touch /var/tmp/smbios_test_file
"""
SMBIOS_SEED_DIR = "/smbios_seed"


def setup_nocloud_local_serial(instance: LXDInstance):
    subp(
        [
            "lxc",
            "config",
            "set",
            instance.name,
            "raw.qemu=-smbios "
            f"type=1,serial=ds=nocloud;s=file://{SMBIOS_SEED_DIR};h=myhost",
        ]
    )


def setup_nocloud_network_serial(instance: LXDInstance):
    subp(
        [
            "lxc",
            "config",
            "set",
            instance.name,
            "raw.qemu=-smbios "
            "type=1,serial=ds=nocloud-net;s=http://0.0.0.0/;h=myhost",
        ]
    )


@pytest.mark.lxd_use_exec
@pytest.mark.skipif(
    PLATFORM != "lxd_vm",
    reason="Requires NoCloud with raw QEMU serial setup",
)
class TestSmbios:
    @pytest.mark.lxd_setup.with_args(setup_nocloud_local_serial)
    def test_smbios_seed_local(self, client: IntegrationInstance):
        """Check that smbios seeds that use local disk work"""
        assert client.execute(f"mkdir -p {SMBIOS_SEED_DIR}").ok
        client.write_to_file(f"{SMBIOS_SEED_DIR}/user-data", SMBIOS_USERDATA)
        client.write_to_file(f"{SMBIOS_SEED_DIR}/meta-data", "")
        client.write_to_file(f"{SMBIOS_SEED_DIR}/vendor-data", "")
        assert client.execute("cloud-init clean --logs").ok
        client.restart()
        assert client.execute("test -f /var/tmp/smbios_test_file").ok

    @pytest.mark.lxd_setup.with_args(setup_nocloud_network_serial)
    def test_smbios_seed_network(self, client: IntegrationInstance):
        """Check that smbios seeds that use network (http/https) work"""
        service_file = "/lib/systemd/system/local-server.service"
        client.write_to_file(
            service_file,
            dedent(
                """\
                [Unit]
                Description=Serve a local webserver
                Before=cloud-init.service
                Wants=cloud-init-local.service
                DefaultDependencies=no
                After=systemd-networkd-wait-online.service
                After=networking.service


                [Install]
                WantedBy=cloud-init.target

                [Service]
                """
                f"WorkingDirectory={SMBIOS_SEED_DIR}"
                """
                ExecStart=/usr/bin/env python3 -m http.server --bind 0.0.0.0 80
                """
            ),
        )
        assert client.execute(
            "chmod 644 /lib/systemd/system/local-server.service"
        ).ok
        assert client.execute("systemctl enable local-server.service").ok
        client.write_to_file(
            "/etc/cloud/cloud.cfg.d/91_do_not_use_lxd.cfg",
            "datasource_list: [ NoCloud, None ]\n",
        )
        assert client.execute(f"mkdir -p {SMBIOS_SEED_DIR}").ok
        client.write_to_file(f"{SMBIOS_SEED_DIR}/user-data", SMBIOS_USERDATA)
        client.write_to_file(f"{SMBIOS_SEED_DIR}/meta-data", "")
        client.write_to_file(f"{SMBIOS_SEED_DIR}/vendor-data", "")
        assert client.execute("cloud-init clean --logs").ok
        client.restart()
        assert client.execute("test -f /var/tmp/smbios_test_file").ok
        assert "'nocloud-net' datasource name is deprecated" in client.execute(
            "cloud-init status --format json"
        )



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

import pytest

from cloudinit.net.ephemeral import EphemeralIPNetwork
from cloudinit.subp import ProcessExecutionError
from tests.unittests.helpers import does_not_raise
from tests.unittests.util import MockDistro

M_PATH = "cloudinit.net.ephemeral."


class TestEphemeralIPNetwork:
    @pytest.mark.parametrize("ipv6", [False, True])
    @pytest.mark.parametrize("ipv4", [False, True])
    @mock.patch(M_PATH + "contextlib.ExitStack")
    @mock.patch(M_PATH + "EphemeralIPv6Network")
    @mock.patch(M_PATH + "EphemeralDHCPv4")
    def test_stack_order(
        self,
        m_ephemeral_dhcp_v4,
        m_ephemeral_ip_v6_network,
        m_exit_stack,
        ipv4,
        ipv6,
    ):
        interface = object()
        distro = MockDistro()
        with EphemeralIPNetwork(distro, interface, ipv4=ipv4, ipv6=ipv6):
            pass
        expected_call_args_list = []
        if ipv4:
            expected_call_args_list.append(
                mock.call(m_ephemeral_dhcp_v4.return_value)
            )
            assert [
                mock.call(distro, interface)
            ] == m_ephemeral_dhcp_v4.call_args_list
        else:
            assert [] == m_ephemeral_dhcp_v4.call_args_list
        if ipv6:
            expected_call_args_list.append(
                mock.call(m_ephemeral_ip_v6_network.return_value)
            )
            assert [
                mock.call(distro, interface)
            ] == m_ephemeral_ip_v6_network.call_args_list
        else:
            assert [] == m_ephemeral_ip_v6_network.call_args_list
        assert (
            expected_call_args_list
            == m_exit_stack.return_value.enter_context.call_args_list
        )

    @pytest.mark.parametrize(
        "m_v4, m_v6, m_context, m_side_effects",
        [
            pytest.param(
                False, True, does_not_raise(), [None, None], id="v6_only"
            ),
            pytest.param(
                True, False, does_not_raise(), [None, None], id="v4_only"
            ),
            pytest.param(
                True,
                True,
                does_not_raise(),
                [ProcessExecutionError, None],
                id="v4_error",
            ),
            pytest.param(
                True,
                True,
                does_not_raise(),
                [None, ProcessExecutionError],
                id="v6_error",
            ),
            pytest.param(
                True,
                True,
                pytest.raises(ProcessExecutionError),
                [
                    ProcessExecutionError,
                    ProcessExecutionError,
                ],
                id="v4_v6_error",
            ),
        ],
    )
    def test_interface_init_failures(
        self, m_v4, m_v6, m_context, m_side_effects, mocker
    ):
        mocker.patch(
            "cloudinit.net.ephemeral.EphemeralDHCPv4"
        ).return_value.__enter__.side_effect = m_side_effects[0]
        mocker.patch(
            "cloudinit.net.ephemeral.EphemeralIPv6Network"
        ).return_value.__enter__.side_effect = m_side_effects[1]
        distro = MockDistro()
        with m_context:
            with EphemeralIPNetwork(distro, "eth0", ipv4=m_v4, ipv6=m_v6):
                pass



# This file is part of cloud-init. See LICENSE file for license information.
import logging

import pytest

from cloudinit.config import cc_landscape
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import mock, skipUnlessJsonSchema, wrap_and_call
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


@mock.patch("cloudinit.config.cc_landscape.subp.subp")
class TestLandscape:
    def test_skip_empty_landscape_cloudconfig(self, m_subp):
        """Empty landscape cloud-config section does no work."""
        mycloud = get_cloud()
        mycloud.distro = mock.MagicMock()
        cfg = {"landscape": {}}
        cc_landscape.handle("notimportant", cfg, mycloud, None)
        assert mycloud.distro.install_packages.called is False

    def test_handler_error_on_invalid_landscape_type(self, m_subp):
        """Raise an error when landscape configuraiton option is invalid."""
        mycloud = get_cloud("ubuntu")
        cfg = {"landscape": "wrongtype"}
        with pytest.raises(RuntimeError) as exc:
            cc_landscape.handle("notimportant", cfg, mycloud, None)
        assert "'landscape' key existed in config, but not a dict" in str(
            exc.value
        )

    def test_handler_restarts_landscape_client(self, m_subp, tmpdir):
        """handler restarts landscape-client after install."""
        mycloud = get_cloud("ubuntu")
        mycloud.distro = mock.MagicMock()
        cfg = {"landscape": {"client": {}}}
        default_fn = tmpdir.join("default")
        wrap_and_call(
            "cloudinit.config.cc_landscape",
            {
                "LSC_CLIENT_CFG_FILE": {
                    "new": tmpdir.join("client.conf").strpath
                },
                "LS_DEFAULT_FILE": {"new": default_fn.strpath},
            },
            cc_landscape.handle,
            "notimportant",
            cfg,
            mycloud,
            None,
        )
        mycloud.distro.install_packages.assert_called_once_with(
            ["landscape-client"]
        )
        assert [
            mock.call(
                [
                    "landscape-config",
                    "--silent",
                    "--data-path",
                    "/var/lib/landscape/client",
                    "--log-level",
                    "info",
                    "--ping-url",
                    "http://landscape.canonical.com/ping",
                    "--url",
                    "https://landscape.canonical.com/message-system",
                ]
            ),
        ] == m_subp.call_args_list

    def test_handler_installs_client_from_ppa_and_supports_overrides(
        self, m_subp, tmpdir
    ):
        """Call landscape-config with any filesystem overrides."""
        mycloud = get_cloud("ubuntu")
        mycloud.distro = mock.MagicMock()
        default_fn = tmpdir.join("default")
        client_fn = tmpdir.join("client.conf")
        client_fn.write("[client]\ndata_path = /var/lib/data\n")
        cfg = {
            "landscape": {
                "install_source": "ppa:landscape/self-hosted-beta",
                "client": {},
            }
        }
        expected_calls = [
            mock.call(
                [
                    "landscape-config",
                    "--silent",
                    "--data-path",
                    "/var/lib/data",
                    "--log-level",
                    "info",
                    "--ping-url",
                    "http://landscape.canonical.com/ping",
                    "--url",
                    "https://landscape.canonical.com/message-system",
                ]
            ),
        ]
        wrap_and_call(
            "cloudinit.config.cc_landscape",
            {
                "LSC_CLIENT_CFG_FILE": {"new": client_fn.strpath},
                "LS_DEFAULT_FILE": {"new": default_fn.strpath},
            },
            cc_landscape.handle,
            "notimportant",
            cfg,
            mycloud,
            None,
        )
        mycloud.distro.install_packages.assert_called_once_with(
            ["landscape-client"]
        )
        assert expected_calls == m_subp.call_args_list
        assert "RUN=1\n" == default_fn.read()

    def test_handler_writes_merged_client_config_file_with_defaults(
        self, m_subp, tmpdir
    ):
        """Merge and write options from LSC_CLIENT_CFG_FILE with defaults."""
        # Write existing sparse client.conf file
        client_fn = tmpdir.join("client.conf")
        client_fn.write("[client]\ncomputer_title = My PC\n")
        default_fn = tmpdir.join("default")
        mycloud = get_cloud("ubuntu")
        mycloud.distro = mock.MagicMock()
        cfg = {"landscape": {"client": {}}}
        expected_calls = [
            mock.call(
                [
                    "landscape-config",
                    "--silent",
                    "--computer-title",
                    "My PC",
                    "--data-path",
                    "/var/lib/landscape/client",
                    "--log-level",
                    "info",
                    "--ping-url",
                    "http://landscape.canonical.com/ping",
                    "--url",
                    "https://landscape.canonical.com/message-system",
                ]
            ),
        ]
        wrap_and_call(
            "cloudinit.config.cc_landscape",
            {
                "LSC_CLIENT_CFG_FILE": {"new": client_fn.strpath},
                "LS_DEFAULT_FILE": {"new": default_fn.strpath},
            },
            cc_landscape.handle,
            "notimportant",
            cfg,
            mycloud,
            None,
        )
        assert expected_calls == m_subp.call_args_list

    def test_handler_writes_merged_provided_cloudconfig_with_defaults(
        self, m_subp, tmpdir
    ):
        """Merge and write options from cloud-config options with defaults."""
        # Write empty sparse client.conf file
        client_fn = tmpdir.join("client.conf")
        client_fn.write("")
        default_fn = tmpdir.join("default")
        mycloud = get_cloud("ubuntu")
        mycloud.distro = mock.MagicMock()
        cfg = {"landscape": {"client": {"computer_title": 'My" PC'}}}
        expected_calls = [
            mock.call(
                [
                    "landscape-config",
                    "--silent",
                    "--computer-title",
                    'My" PC',
                    "--data-path",
                    "/var/lib/landscape/client",
                    "--log-level",
                    "info",
                    "--ping-url",
                    "http://landscape.canonical.com/ping",
                    "--url",
                    "https://landscape.canonical.com/message-system",
                ]
            ),
        ]
        wrap_and_call(
            "cloudinit.config.cc_landscape",
            {
                "LSC_CLIENT_CFG_FILE": {"new": client_fn.strpath},
                "LS_DEFAULT_FILE": {"new": default_fn.strpath},
            },
            cc_landscape.handle,
            "notimportant",
            cfg,
            mycloud,
            None,
        )
        assert expected_calls == m_subp.call_args_list


class TestLandscapeSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Allow undocumented keys client keys without error
            (
                {
                    "landscape": {
                        "client": {
                            "computer_title": "joe",
                            "account_name": "joe's acct",
                            "allow_additional_keys": 1,
                        }
                    }
                },
                None,
            ),
            (
                {
                    "landscape": {
                        "client": {
                            "computer_title": "joe",
                            "account_name": "joe's acct",
                        }
                    }
                },
                None,
            ),
            # tags are comma-delimited
            (
                {
                    "landscape": {
                        "client": {
                            "computer_title": "joe",
                            "account_name": "joe's acct",
                            "tags": "1,2,3",
                        }
                    }
                },
                None,
            ),
            (
                {
                    "landscape": {
                        "client": {
                            "computer_title": "joe",
                            "account_name": "joe's acct",
                            "client": {"tags": "1"},
                        }
                    }
                },
                None,
            ),
            (
                {
                    "landscape": {
                        "client": {
                            "computer_title": "joe",
                            "account_name": "joe's acct",
                        },
                        "random-config-value": {"tags": "1"},
                    }
                },
                "Additional properties are not allowed",
            ),
            # Require client key
            ({"landscape": {}}, "'client' is a required property"),
            # Require client.account_name and client.computer_title
            (
                {"landscape": {"client": {"computer_title": "joe"}}},
                "'account_name' is a required property",
            ),
            (
                {"landscape": {"client": {"account_name": "joe's acct"}}},
                "'computer_title' is a required property",
            ),
            # tags are not whitespace-delimited
            (
                {"landscape": {"client": {"tags": "1, 2,3"}}},
                "'1, 2,3' does not match",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit.distros.parsers.ifconfig import Ifconfig
from tests.unittests.helpers import TestCase, readResource


class TestIfconfigParserFreeBSD(TestCase):
    def setUp(self):
        super(TestIfconfigParserFreeBSD, self).setUp()
        self.ifs_txt = readResource("netinfo/freebsd-ifconfig-output")

    def test_parse_freebsd(self):
        """assert parsing works without any exceptions"""
        Ifconfig().parse(self.ifs_txt)

    def test_is_bridge(self):
        """assert bridge0 is_bridge"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["bridge0"].is_bridge

    def test_index(self):
        """assert vtnet0 index is 1"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["vtnet0"].index == 1

    def test_is_vlan(self):
        """assert re0.33 is_vlan"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["re0.33"].is_vlan

    def test_description(self):
        """assert vnet0:11 is associated with jail: webirc"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["vnet0:11"].description == "'associated with jail: webirc'"

    def test_vtnet_options(self):
        """assert vtnet has TXCSUM"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert "txcsum" in ifs["vtnet0"].options

    def test_duplicate_mac(self):
        """
        assert that we can have duplicate macs, and that it's not an accident
        """
        self.ifs_txt = readResource(
            "netinfo/freebsd-duplicate-macs-ifconfig-output"
        )
        ifc = Ifconfig()
        ifc.parse(self.ifs_txt)
        ifs_by_mac = ifc.ifs_by_mac()
        assert len(ifs_by_mac["00:0d:3a:54:ad:1e"]) == 2
        assert (
            ifs_by_mac["00:0d:3a:54:ad:1e"][0].name
            != ifs_by_mac["00:0d:3a:54:ad:1e"][1].name
        )


class TestIfconfigParserOpenBSD(TestCase):
    def setUp(self):
        super(TestIfconfigParserOpenBSD, self).setUp()
        self.ifs_txt = readResource("netinfo/openbsd-ifconfig-output")

    def test_parse_openbsd(self):
        """assert parsing works without any exceptions"""
        Ifconfig().parse(self.ifs_txt)

    def test_is_not_physical(self):
        """assert enc0 is not physical"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert not ifs["enc0"].is_physical

    def test_is_physical(self):
        """assert enc0 is not physical"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["vio0"].is_physical

    def test_index(self):
        """assert vio0 index is 1"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["vio0"].index == 1

    def test_gif_ipv6(self):
        """assert that we can parse a gif inet6 address, despite the -->"""
        ifs = Ifconfig().parse(self.ifs_txt)
        assert ifs["gif0"].inet6["fe80::be30:5bff:fed0:471"] == {
            "prefixlen": "64",
            "scope": "link-local",
        }



"""Integration test for the ssh module.

This module has two tests to verify if we can create ssh keys
through the ``ssh`` module. The first test asserts that some keys
were not created while the second one verifies if the expected
keys were created.

(This is ported from
``tests/cloud_tests/testcases/modules/ssh_keys_generate.yaml``.)"""

import pytest

USER_DATA = """\
#cloud-config
ssh_genkeytypes:
  - ecdsa
  - ed25519
authkey_hash: sha512
"""


@pytest.mark.ci
@pytest.mark.user_data(USER_DATA)
class TestSshKeysGenerate:
    @pytest.mark.parametrize(
        "ssh_key_path",
        (
            "/etc/ssh/ssh_host_rsa_key.pub",
            "/etc/ssh/ssh_host_rsa_key",
        ),
    )
    def test_ssh_keys_not_generated(self, ssh_key_path, class_client):
        out = class_client.execute("test -e {}".format(ssh_key_path))
        assert out.failed

    @pytest.mark.parametrize(
        "ssh_key_path",
        (
            "/etc/ssh/ssh_host_ecdsa_key.pub",
            "/etc/ssh/ssh_host_ecdsa_key",
            "/etc/ssh/ssh_host_ed25519_key.pub",
            "/etc/ssh/ssh_host_ed25519_key",
        ),
    )
    def test_ssh_keys_generated(self, ssh_key_path, class_client):
        out = class_client.read_from_file(ssh_key_path)
        assert "" != out.strip()



# This file is part of cloud-init. See LICENSE file for license information.
import logging
from unittest import mock

import pytest

from cloudinit import subp
from cloudinit.config.cc_package_update_upgrade_install import (
    REBOOT_FILES,
    handle,
)
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.distros import PackageInstallerError
from cloudinit.subp import SubpResult
from tests.unittests.helpers import (
    SCHEMA_EMPTY_ERROR,
    does_not_raise,
    skipUnlessJsonSchema,
)
from tests.unittests.util import get_cloud

M_PATH = "cloudinit.config.cc_package_update_upgrade_install."


@pytest.fixture
def common_mocks(mocker):
    mocker.patch("os.path.isfile", return_value=False)
    mocker.patch("cloudinit.distros.Distro.update_package_sources")
    mocker.patch(
        "cloudinit.distros.package_management.apt.Apt.update_package_sources"
    )
    mocker.patch(
        "cloudinit.distros.package_management.apt.Apt._apt_lock_available",
        return_value=True,
    )
    mocker.patch(
        "cloudinit.distros.package_management.apt.Apt.available",
        return_value=True,
    )
    mocker.patch(
        "cloudinit.distros.package_management.snap.Snap.available",
        return_value=True,
    )


class TestRebootIfRequired:
    @pytest.mark.parametrize(
        "cloud_cfg,reboot_file,expectation",
        (
            pytest.param(
                {"package_reboot_if_required": True},
                "/run/reboot-needed",
                does_not_raise(),
                id="no_reboot_when_no_package_changes",
            ),
            pytest.param(
                {"package_reboot_if_required": True, "package_upgrade": True},
                "/run/reboot-needed",
                pytest.raises(
                    RuntimeError
                ),  # _fire_reboot raises RuntimeError
                id="perform_reboot_on_package_upgrade_and_suse_reboot_marker",
            ),
            pytest.param(
                {"package_reboot_if_required": True, "package_upgrade": True},
                "",  # No reboot-needed flag file present
                does_not_raise(),
                id="no_reboot_on_package_upgrade_and_no_reboot_required_file",
            ),
            pytest.param(
                {"package_reboot_if_required": True, "package_upgrade": True},
                "/var/run/reboot-required",
                pytest.raises(
                    RuntimeError
                ),  # _fire_reboot raises RuntimeError
                id="perform_reboot_on_package_upgrade_and_reboot_marker",
            ),
            pytest.param(
                {"package_reboot_if_required": True, "packages": ["sl"]},
                "/var/run/reboot-required",
                pytest.raises(
                    RuntimeError
                ),  # _fire_reboot raises RuntimeError
                id="perform_reboot_on_packages_and_reboot_marker",
            ),
        ),
    )
    def test_wb_only_reboot_on_reboot_when_configured_and_required(
        self, cloud_cfg, reboot_file, expectation, common_mocks, caplog
    ):
        """Only reboot when packages are updated and reboot_if_required.

        Whitebox testing because _fire_reboot will not actually reboot the
        system and we expect to fallback to a raised RuntimeError in testing

        NOOP when any of the following are not true:
          - no reboot_if_requred: true config
          - no reboot-required flag files exist
          - no packages were changed by cloud-init via upgrade or packages cfg
        """

        def _isfile(filename: str):
            return filename == reboot_file

        cloud = get_cloud("ubuntu")

        subp_call = None
        sleep_count = 0
        if cloud_cfg.get("package_reboot_if_required"):
            if reboot_file in REBOOT_FILES:
                if cloud_cfg.get("package_upgrade") or cloud_cfg.get(
                    "packages"
                ):
                    sleep_count = 6
                    # Expect a RuntimeError after sleeps because of mocked
                    # subp and not really rebooting the system
                    subp_call = ["/sbin/reboot"]

        caplog.set_level(logging.WARNING)
        with mock.patch(
            "cloudinit.subp.subp", return_value=("fakeout", "fakeerr")
        ) as m_subp:
            with mock.patch("os.path.isfile", side_effect=_isfile):
                with mock.patch(M_PATH + "time.sleep") as m_sleep:
                    with mock.patch(M_PATH + "flush_loggers"):
                        with expectation:
                            handle("", cloud_cfg, cloud, [])
        assert sleep_count == m_sleep.call_count
        if subp_call:
            assert (
                f"Rebooting after upgrade or install per {reboot_file}"
                in caplog.text
            )
            m_subp.assert_called_with(subp_call)


class TestMultiplePackageManagers:
    def test_explicit_apt(self, common_mocks):
        def _new_subp(*args, **kwargs):
            if args and "apt-cache" in args[0]:
                return SubpResult("pkg1\npkg2\npkg3", None)

        cloud = get_cloud("ubuntu")
        cfg = {"packages": [{"apt": ["pkg1", "pkg2"]}]}
        with mock.patch(
            "cloudinit.subp.subp", side_effect=_new_subp
        ) as m_subp:
            handle("", cfg, cloud, [])

        assert len(m_subp.call_args_list) == 2
        assert m_subp.call_args_list[0] == mock.call(["apt-cache", "pkgnames"])

        for arg in ["apt-get", "install", "pkg1", "pkg2"]:
            assert arg in m_subp.call_args_list[1][1]["args"]

    def test_explicit_apt_version(self, common_mocks):
        def _new_subp(*args, **kwargs):
            if args and "apt-cache" in args[0]:
                return SubpResult("pkg1\npkg2\npkg3", None)

        cloud = get_cloud("ubuntu")
        cfg = {"packages": [{"apt": ["pkg1", ["pkg2", "1.2.3"]]}]}
        with mock.patch(
            "cloudinit.subp.subp", side_effect=_new_subp
        ) as m_subp:
            handle("", cfg, cloud, [])

        assert len(m_subp.call_args_list) == 2
        assert m_subp.call_args_list[0] == mock.call(["apt-cache", "pkgnames"])

        for arg in ["apt-get", "install", "pkg1", "pkg2=1.2.3"]:
            assert arg in m_subp.call_args_list[1][1]["args"]

    @mock.patch("cloudinit.subp.subp")
    def test_explicit_snap(self, m_subp, common_mocks):
        cloud = get_cloud("ubuntu")
        cfg = {"packages": [{"snap": ["pkg1", "pkg2"]}]}
        handle("", cfg, cloud, [])

        assert len(m_subp.call_args_list) == 2
        assert mock.call(["snap", "install", "pkg1"]) in m_subp.call_args_list
        assert mock.call(["snap", "install", "pkg2"]) in m_subp.call_args_list

    @mock.patch("cloudinit.subp.subp")
    def test_explicit_snap_version(self, m_subp, common_mocks):
        cloud = get_cloud("ubuntu")
        cfg = {"packages": [{"snap": ["pkg1", ["pkg2", "--edge"]]}]}
        handle("", cfg, cloud, [])

        assert len(m_subp.call_args_list) == 2
        assert mock.call(["snap", "install", "pkg1"]) in m_subp.call_args_list
        assert (
            mock.call(["snap", "install", "pkg2", "--edge"])
            in m_subp.call_args_list
        )

    def test_combined(self, common_mocks):
        """Ensure that pkg1 is installed by snap since it isn't available
        in the apt cache, and ensure pkg5 is installed by snap even though it
        is available under apt because it is explicitly listed under snap."""

        def _new_subp(*args, **kwargs):
            if args and "apt-cache" in args[0]:
                return SubpResult("pkg2\npkg3\npkg5\npkg6", None)

        cloud = get_cloud("ubuntu")
        cfg = {
            "packages": [
                "pkg1",
                {"apt": ["pkg2", ["pkg3", "1.2.3"]]},
                {"snap": ["pkg4", ["pkg5", "--edge"]]},
                "pkg6",
            ],
        }
        with mock.patch(
            "cloudinit.subp.subp", side_effect=_new_subp
        ) as m_subp:
            handle("", cfg, cloud, [])

        assert len(m_subp.call_args_list) == 5
        assert m_subp.call_args_list[0] == mock.call(["apt-cache", "pkgnames"])
        for arg in ["apt-get", "install", "pkg2", "pkg3=1.2.3", "pkg6"]:
            assert arg in m_subp.call_args_list[1][1]["args"]

        assert mock.call(["snap", "install", "pkg1"]) in m_subp.call_args_list
        assert (
            mock.call(["snap", "install", "pkg5", "--edge"])
            in m_subp.call_args_list
        )
        assert mock.call(["snap", "install", "pkg4"]) in m_subp.call_args_list

    def test_error_apt(self, common_mocks):
        """Since we have already checked that the package(s) exists in apt,
        if we have an error in apt, we don't want to fall through to
        additional package installs.
        """

        def _new_subp(*args, **kwargs):
            if args and "apt-cache" in args[0]:
                return SubpResult("pkg1", None)
            if "args" in kwargs and "install" in kwargs["args"]:
                raise subp.ProcessExecutionError(
                    cmd=kwargs["args"],
                    stdout="dontcare",
                    stderr="E: Unable to locate package pkg1",
                    exit_code=100,
                )

        cloud = get_cloud("ubuntu")
        cfg = {"packages": ["pkg1"]}
        with mock.patch("cloudinit.subp.subp", side_effect=_new_subp):
            with pytest.raises(subp.ProcessExecutionError):
                handle("", cfg, cloud, [])

    def test_error_snap(self, common_mocks, caplog):
        """Since we haven't checked the package(s) existence, we should fall
        through to additional package installs.
        """
        caplog.set_level(logging.DEBUG)

        def _new_subp(*args, **kwargs):
            if args:
                if "apt-cache" in args[0]:
                    return SubpResult("", None)
                if "install" in args[0]:
                    raise subp.ProcessExecutionError(
                        cmd=args[0],
                        stdout="dontcare",
                        stderr='error: snap "pkg1" not found',
                        exit_code=1,
                    )

        cloud = get_cloud("ubuntu")
        cfg = {"packages": ["pkg1"]}
        with mock.patch("cloudinit.subp.subp", side_effect=_new_subp):
            with pytest.raises(PackageInstallerError):
                handle("", cfg, cloud, [])

        assert caplog.records[-3].levelname == "WARNING"
        assert (
            caplog.records[-3].message
            == "Failure when attempting to install packages: ['pkg1']"
        )


class TestPackageUpdateUpgradeSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # packages list with single entry (2 required)
            ({"packages": ["p1", ["p2"]]}, ""),
            # packages list with three entries (2 required)
            ({"packages": ["p1", ["p2", "p3", "p4"]]}, ""),
            # empty packages list
            ({"packages": []}, SCHEMA_EMPTY_ERROR),
            (
                {"apt_update": False},
                (
                    "Cloud config schema deprecations: apt_update: "
                    "Default: ``false``. Deprecated in version 22.2. "
                    "Use ``package_update`` instead."
                ),
            ),
            (
                {"apt_upgrade": False},
                (
                    "Cloud config schema deprecations: apt_upgrade: "
                    "Default: ``false``. Deprecated in version 22.2. "
                    "Use ``package_upgrade`` instead."
                ),
            ),
            (
                {"apt_reboot_if_required": False},
                (
                    "Cloud config schema deprecations: "
                    "apt_reboot_if_required: Default: ``false``. "
                    "Deprecated in version 22.2. Use "
                    "``package_reboot_if_required`` instead."
                ),
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        with pytest.raises(SchemaValidationError, match=error_msg):
            validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
import logging
import os
import stat
from unittest.mock import patch

import pytest

from cloudinit import helpers, subp, util
from cloudinit.config.cc_runcmd import handle
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import (
    SCHEMA_EMPTY_ERROR,
    FilesystemMockingTestCase,
    skipUnlessJsonSchema,
)
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


class TestRuncmd(FilesystemMockingTestCase):

    with_logs = True

    def setUp(self):
        super(TestRuncmd, self).setUp()
        self.subp = subp.subp
        self.new_root = self.tmp_dir()
        self.patchUtils(self.new_root)
        self.paths = helpers.Paths({"scripts": self.new_root})

    def test_handler_skip_if_no_runcmd(self):
        """When the provided config doesn't contain runcmd, skip it."""
        cfg = {}
        mycloud = get_cloud(paths=self.paths)
        handle("notimportant", cfg, mycloud, None)
        self.assertIn(
            "Skipping module named notimportant, no 'runcmd' key",
            self.logs.getvalue(),
        )

    @patch("cloudinit.util.shellify")
    def test_runcmd_shellify_fails(self, cls):
        """When shellify fails throw exception"""
        cls.side_effect = TypeError("patched shellify")
        valid_config = {"runcmd": ["echo 42"]}
        cc = get_cloud(paths=self.paths)
        with self.assertRaises(TypeError) as cm:
            with self.allow_subp(["/bin/sh"]):
                handle("cc_runcmd", valid_config, cc, None)
        self.assertIn("Failed to shellify", str(cm.exception))

    def test_handler_invalid_command_set(self):
        """Commands which can't be converted to shell will raise errors."""
        invalid_config = {"runcmd": 1}
        cc = get_cloud(paths=self.paths)
        with self.assertRaises(TypeError) as cm:
            handle("cc_runcmd", invalid_config, cc, [])
        self.assertIn(
            "Failed to shellify 1 into file"
            " /var/lib/cloud/instances/iid-datasource-none/scripts/runcmd",
            str(cm.exception),
        )

    def test_handler_write_valid_runcmd_schema_to_file(self):
        """Valid runcmd schema is written to a runcmd shell script."""
        valid_config = {"runcmd": [["ls", "/"]]}
        cc = get_cloud(paths=self.paths)
        handle("cc_runcmd", valid_config, cc, [])
        runcmd_file = os.path.join(
            self.new_root,
            "var/lib/cloud/instances/iid-datasource-none/scripts/runcmd",
        )
        self.assertEqual(
            "#!/bin/sh\n'ls' '/'\n", util.load_text_file(runcmd_file)
        )
        file_stat = os.stat(runcmd_file)
        self.assertEqual(0o700, stat.S_IMODE(file_stat.st_mode))


@skipUnlessJsonSchema()
class TestRunCmdSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        (
            # Ensure duplicate commands are valid
            ({"runcmd": [["echo", "bye"], ["echo", "bye"]]}, None),
            ({"runcmd": ["echo bye", "echo bye"]}, None),
            # Invalid schemas
            ({"runcmd": 1}, "1 is not of type 'array'"),
            ({"runcmd": []}, rf"runcmd: \[\] {SCHEMA_EMPTY_ERROR}"),
            (
                {
                    "runcmd": [
                        "ls /",
                        20,
                        ["wget", "http://stuff/blah"],
                        {"a": "n"},
                    ]
                },
                "",
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
"""Common utility functions for interacting with subprocess."""

import collections
import logging
import os
import subprocess
import time
from errno import ENOEXEC
from io import TextIOWrapper
from typing import List, Union

LOG = logging.getLogger(__name__)

SubpResult = collections.namedtuple("SubpResult", ["stdout", "stderr"])


def prepend_base_command(base_command, commands):
    """Ensure user-provided commands start with base_command; warn otherwise.

    Each command is either a list or string. Perform the following:
       - If the command is a list, pop the first element if it is None
       - If the command is a list, insert base_command as the first element if
         not present.
       - When the command is a string not starting with 'base-command', warn.

    Allow flexibility to provide non-base-command environment/config setup if
    needed.

    @commands: List of commands. Each command element is a list or string.

    @return: List of 'fixed up' commands.
    @raise: TypeError on invalid config item type.
    """
    warnings = []
    errors = []
    fixed_commands = []
    for command in commands:
        if isinstance(command, list):
            if command[0] is None:  # Avoid warnings by specifying None
                command = command[1:]
            elif command[0] != base_command:  # Automatically prepend
                command.insert(0, base_command)
        elif isinstance(command, str):
            if not command.startswith(f"{base_command} "):
                warnings.append(command)
        else:
            errors.append(str(command))
            continue
        fixed_commands.append(command)

    if warnings:
        LOG.warning(
            "Non-%s commands in %s config:\n%s",
            base_command,
            base_command,
            "\n".join(warnings),
        )
    if errors:
        raise TypeError(
            "Invalid {name} config."
            " These commands are not a string or list:\n{errors}".format(
                name=base_command, errors="\n".join(errors)
            )
        )
    return fixed_commands


class ProcessExecutionError(IOError):
    MESSAGE_TMPL = (
        "%(description)s\n"
        "Command: %(cmd)s\n"
        "Exit code: %(exit_code)s\n"
        "Reason: %(reason)s\n"
        "Stdout: %(stdout)s\n"
        "Stderr: %(stderr)s"
    )
    empty_attr = "-"

    def __init__(
        self,
        stdout=None,
        stderr=None,
        exit_code=None,
        cmd=None,
        description=None,
        reason=None,
        errno=None,
    ):
        self.cmd = cmd or self.empty_attr

        if description:
            self.description = description
        elif not exit_code and errno == ENOEXEC:
            self.description = "Exec format error. Missing #! in script?"
        else:
            self.description = "Unexpected error while running command."

        self.exit_code = (
            exit_code if isinstance(exit_code, int) else self.empty_attr
        )

        if not stderr:
            self.stderr = self.empty_attr if stderr is None else stderr
        else:
            self.stderr = self._indent_text(stderr)

        if not stdout:
            self.stdout = self.empty_attr if stdout is None else stdout
        else:
            self.stdout = self._indent_text(stdout)

        self.reason = reason or self.empty_attr

        if errno:
            self.errno = errno
        message = self.MESSAGE_TMPL % {
            "description": self._ensure_string(self.description),
            "cmd": self._ensure_string(self.cmd),
            "exit_code": self._ensure_string(self.exit_code),
            "stdout": self._ensure_string(self.stdout),
            "stderr": self._ensure_string(self.stderr),
            "reason": self._ensure_string(self.reason),
        }
        IOError.__init__(self, message)

    def _ensure_string(self, text):
        """
        if data is bytes object, decode
        """
        return text.decode() if isinstance(text, bytes) else text

    def _indent_text(
        self, text: Union[str, bytes], indent_level=8
    ) -> Union[str, bytes]:
        """
        indent text on all but the first line, allowing for easy to read output

        remove any newlines at end of text first to prevent unneeded blank
        line in output
        """
        if not isinstance(text, bytes):
            return text.rstrip("\n").replace("\n", "\n" + " " * indent_level)
        return text.rstrip(b"\n").replace(b"\n", b"\n" + b" " * indent_level)


def raise_on_invalid_command(args: Union[List[str], List[bytes]]):
    """check argument types to ensure that subp() can run the argument

    Throw a user-friendly exception which explains the issue.

    args: list of arguments passed to subp()
    raises: ProcessExecutionError with information explaining the issue
    """
    for component in args:
        # if already bytes, or implements encode(), then it should be safe
        if not (isinstance(component, bytes) or hasattr(component, "encode")):
            LOG.warning("Running invalid command: %s", args)
            raise ProcessExecutionError(
                cmd=args, reason=f"Running invalid command: {args}"
            )


def subp(
    args: Union[str, bytes, List[str], List[bytes]],
    *,
    data=None,
    rcs=None,
    capture=True,
    shell=False,
    logstring=False,
    decode="replace",
    update_env=None,
    cwd=None,
    timeout=None,
) -> SubpResult:
    """Run a subprocess.

    :param args: command to run in a list. [cmd, arg1, arg2...]
    :param data: input to the command, made available on its stdin.
    :param rcs:
        a list of allowed return codes.  If subprocess exits with a value not
        in this list, a ProcessExecutionError will be raised.  By default,
        data is returned as a string.  See 'decode' parameter.
    :param capture:
        boolean indicating if output should be captured.  If True, then stderr
        and stdout will be returned.  If False, they will not be redirected.
    :param shell: boolean indicating if this should be run with a shell.
    :param logstring:
        the command will be logged to DEBUG.  If it contains info that should
        not be logged, then logstring will be logged instead.
    :param decode:
        if False, no decoding will be done and returned stdout and stderr will
        be bytes.  Other allowed values are 'strict', 'ignore', and 'replace'.
        These values are passed through to bytes().decode() as the 'errors'
        parameter.  There is no support for decoding to other than utf-8.
    :param update_env:
        update the environment for this command with this dictionary.
        this will not affect the current processes os.environ.
    :param cwd:
        change the working directory to cwd before executing the command.
    :param timeout: maximum time for the subprocess to run, passed directly to
        the timeout parameter of Popen.communicate()

    :return
        if not capturing, return is (None, None)
        if capturing, stdout and stderr are returned.
            if decode:
                entries in tuple will be string
            if not decode:
                entries in tuple will be bytes
    """

    if rcs is None:
        rcs = [0]

    env = os.environ.copy()
    if update_env:
        env.update(update_env)

    if not logstring:
        LOG.debug(
            "Running command %s with allowed return codes %s"
            " (shell=%s, capture=%s)",
            args,
            rcs,
            shell,
            capture,
        )
    else:
        LOG.debug(
            "Running hidden command to protect sensitive "
            "input/output logstring: %s",
            logstring,
        )

    stdin: Union[TextIOWrapper, int]
    stdout = None
    stderr = None
    if capture:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
    if data is None:
        # using devnull assures any reads get null, rather
        # than possibly waiting on input.
        stdin = subprocess.DEVNULL
    else:
        stdin = subprocess.PIPE
        if not isinstance(data, bytes):
            data = data.encode()

    # Popen converts entries in the arguments array from non-bytes to bytes.
    # When locale is unset it may use ascii for that encoding which can
    # cause UnicodeDecodeErrors. (LP: #1751051)
    bytes_args: Union[bytes, List[bytes]]
    if isinstance(args, bytes):
        bytes_args = args
    elif isinstance(args, str):
        bytes_args = args.encode("utf-8")
    else:
        raise_on_invalid_command(args)
        bytes_args = [
            x if isinstance(x, bytes) else x.encode("utf-8") for x in args
        ]
    try:
        before = time.time()
        sp = subprocess.Popen(
            bytes_args,
            stdout=stdout,
            stderr=stderr,
            stdin=stdin,
            env=env,
            shell=shell,
            cwd=cwd,
        )
        out, err = sp.communicate(data, timeout=timeout)
        total = time.time() - before
        if total > 0.1:
            LOG.debug("command %s took %.3ss to run", args, total)
    except OSError as e:
        raise ProcessExecutionError(
            cmd=args,
            reason=e,
            errno=e.errno,
            stdout="-" if decode else b"-",
            stderr="-" if decode else b"-",
        ) from e
    if decode:

        def ldecode(data, m="utf-8"):
            return data.decode(m, decode) if isinstance(data, bytes) else data

        out = ldecode(out)
        err = ldecode(err)

    rc = sp.returncode
    if rc not in rcs:
        raise ProcessExecutionError(
            stdout=out, stderr=err, exit_code=rc, cmd=args
        )
    return SubpResult(out, err)


def target_path(target=None, path=None):
    # return 'path' inside target, accepting target as None
    if target in (None, ""):
        target = "/"
    elif not isinstance(target, str):
        raise ValueError(f"Unexpected input for target: {target}")
    else:
        target = os.path.abspath(target)
        # abspath("//") returns "//" specifically for 2 slashes.
        if target.startswith("//"):
            target = target[1:]

    if not path:
        return target

    # os.path.join("/etc", "/foo") returns "/foo". Chomp all leading /.
    while len(path) and path[0] == "/":
        path = path[1:]
    return os.path.join(target, path)


def which(program, search=None, target=None):
    target = target_path(target)

    if os.path.sep in program and is_exe(target_path(target, program)):
        # if program had a '/' in it, then do not search PATH
        # 'which' does consider cwd here. (cd / && which bin/ls) = bin/ls
        # so effectively we set cwd to / (or target)
        return program

    if search is None:
        paths = [
            p.strip('"') for p in os.environ.get("PATH", "").split(os.pathsep)
        ]
        search = (
            paths if target == "/" else [p for p in paths if p.startswith("/")]
        )
    # normalize path input
    search = [os.path.abspath(p) for p in search]

    for path in search:
        ppath = os.path.sep.join((path, program))
        if is_exe(target_path(target, ppath)):
            return ppath

    return None


def is_exe(fpath):
    # return boolean indicating if fpath exists and is executable.
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def runparts(dirp, skip_no_exist=True, exe_prefix=None):
    if skip_no_exist and not os.path.isdir(dirp):
        return

    failed = []
    attempted = []

    if exe_prefix is None:
        prefix = []
    elif isinstance(exe_prefix, str):
        prefix = [str(exe_prefix)]
    elif isinstance(exe_prefix, list):
        prefix = exe_prefix
    else:
        raise TypeError("exe_prefix must be None, str, or list")

    for exe_name in sorted(os.listdir(dirp)):
        exe_path = os.path.join(dirp, exe_name)
        if is_exe(exe_path):
            attempted.append(exe_path)
            try:
                subp(prefix + [exe_path], capture=False)
            except ProcessExecutionError as e:
                LOG.debug(e)
                failed.append(exe_name)
        elif os.path.isfile(exe_path):
            LOG.warning(
                "skipping %s as its not executable "
                "or the underlying file system is mounted without "
                "executable permissions.",
                exe_path,
            )
        else:
            LOG.debug("Not executing special file [%s]", exe_path)

    if failed and attempted:
        raise RuntimeError(
            f'Runparts: {len(failed)} failures ({",".join(failed)}) in '
            f"{len(attempted)} attempted commands"
        )



# Copyright (C) 2009 Canonical Ltd.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Soren Hansen <soren@canonical.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init.  See LICENSE file for license information.

# Distutils magic for ec2-init

import atexit
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from glob import glob

import setuptools
from setuptools.command.egg_info import egg_info
from setuptools.command.install import install

# Python-path here is a little unpredictable as setup.py could be run
# from a directory other than the root of the repo, so ensure we can find
# our utils
sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))
# isort: off
from setup_utils import (  # noqa: E402
    get_version,
    in_virtualenv,
    is_f,
    is_generator,
    pkg_config_read,
    read_requires,
)

# isort: on
del sys.path[0]

# pylint: disable=W0402
try:
    from setuptools.errors import DistutilsError
except ImportError:
    from distutils.errors import DistutilsArgError as DistutilsError
# pylint: enable=W0402

RENDERED_TMPD_PREFIX = "RENDERED_TEMPD"
VARIANT = None
PREFIX = None


def render_tmpl(template, mode=None, is_yaml=False):
    """render template into a tmpdir under same dir as setup.py

    This is rendered to a temporary directory under the top level
    directory with the name 'cloud.cfg'.  The reason for not just rendering
    to config/cloud.cfg is for a.) don't want to write over contents
    in that file if user had something there. b.) debuild will complain
    that files are different outside of the debian directory."""

    # newer versions just use install.
    if "install" not in sys.argv:
        return template

    tmpl_ext = ".tmpl"
    # we may get passed a non-template file, just pass it back
    if not template.endswith(tmpl_ext):
        return template

    topdir = os.path.dirname(sys.argv[0])
    tmpd = tempfile.mkdtemp(dir=topdir, prefix=RENDERED_TMPD_PREFIX)
    atexit.register(shutil.rmtree, tmpd)
    bname = os.path.basename(template)
    ename, ext = os.path.splitext(bname)
    if ext == tmpl_ext:
        bname = ename
    fpath = os.path.join(tmpd, bname)
    cmd_variant = []
    cmd_prefix = []
    if VARIANT:
        cmd_variant = ["--variant", VARIANT]
    if PREFIX:
        cmd_prefix = ["--prefix", PREFIX]
    subprocess.run(
        [
            sys.executable,
            "./tools/render-template",
            *(["--is-yaml"] if is_yaml else []),
            *cmd_prefix,
            *cmd_variant,
            *[template, fpath],
        ],
        check=True,
    )
    if mode:
        os.chmod(fpath, mode)
    # return path relative to setup.py
    return os.path.join(os.path.basename(tmpd), bname)


# User can set the variant for template rendering
for a in sys.argv:
    if a.startswith("--distro"):
        idx = sys.argv.index(a)
        if "=" in a:
            _, VARIANT = a.split("=")
            del sys.argv[idx]
        else:
            VARIANT = sys.argv[idx + 1]
            del sys.argv[idx + 1]
            sys.argv.remove("--distro")

# parse PREFIX and pass it on from render_tmpl()
for a in sys.argv:
    if a.startswith("--prefix"):
        idx = sys.argv.index(a)
        if "=" in a:
            _, PREFIX = a.split("=")
        else:
            PREFIX = sys.argv[idx + 1]

INITSYS_FILES = {
    "sysvinit": lambda: [f for f in glob("sysvinit/redhat/*") if is_f(f)],
    "sysvinit_freebsd": lambda: [
        render_tmpl(f, mode=0o755)
        for f in glob("sysvinit/freebsd/*")
        if is_f(f)
    ],
    "sysvinit_netbsd": lambda: [
        render_tmpl(f, mode=0o755)
        for f in glob("sysvinit/netbsd/*")
        if is_f(f)
    ],
    "sysvinit_openbsd": lambda: [
        render_tmpl(f, mode=0o755)
        for f in glob("sysvinit/openbsd/*")
        if is_f(f)
    ],
    "sysvinit_deb": lambda: [f for f in glob("sysvinit/debian/*") if is_f(f)],
    "sysvinit_openrc": lambda: [
        f for f in glob("sysvinit/openrc/*") if is_f(f)
    ],
    "sysvinit_openrc.dep": lambda: ["tools/cloud-init-hotplugd"],
    "systemd": lambda: [
        render_tmpl(f)
        for f in (
            glob("systemd/*.tmpl")
            + glob("systemd/*.service")
            + glob("systemd/*.socket")
            + glob("systemd/*.target")
        )
        if (is_f(f) and not is_generator(f))
    ],
    "systemd.generators": lambda: [
        render_tmpl(f, mode=0o755)
        for f in glob("systemd/*")
        if is_f(f) and is_generator(f)
    ],
}
INITSYS_ROOTS = {
    "sysvinit": "etc/rc.d/init.d",
    "sysvinit_freebsd": "usr/local/etc/rc.d",
    "sysvinit_netbsd": "usr/local/etc/rc.d",
    "sysvinit_openbsd": "etc/rc.d",
    "sysvinit_deb": "etc/init.d",
    "sysvinit_openrc": "etc/init.d",
    "sysvinit_openrc.dep": "usr/lib/cloud-init",
    "systemd": pkg_config_read("systemd", "systemdsystemunitdir"),
    "systemd.generators": pkg_config_read(
        "systemd", "systemdsystemgeneratordir"
    ),
}
INITSYS_TYPES = sorted([f.partition(".")[0] for f in INITSYS_ROOTS.keys()])


# Install everything in the right location and take care of Linux (default) and
# FreeBSD systems.
USR = "usr"
ETC = "etc"
USR_LIB_EXEC = "usr/lib"
LIB = "lib"
if os.uname()[0] in ["FreeBSD", "DragonFly", "OpenBSD"]:
    USR = "usr/local"
    USR_LIB_EXEC = "usr/local/lib"
elif os.path.isfile("/etc/redhat-release"):
    USR_LIB_EXEC = "usr/libexec"
elif os.path.isfile("/etc/system-release-cpe"):
    with open("/etc/system-release-cpe") as f:
        cpe_data = f.read().rstrip().split(":")

        if cpe_data[1] == "\o":  # noqa: W605
            # URI formatted CPE
            inc = 0
        else:
            # String formatted CPE
            inc = 1
        (cpe_vendor, cpe_product, cpe_version) = cpe_data[2 + inc : 5 + inc]
        if cpe_vendor == "amazon":
            USR_LIB_EXEC = "usr/libexec"


class MyEggInfo(egg_info):
    """This makes sure to not include the rendered files in SOURCES.txt."""

    def find_sources(self):
        egg_info.find_sources(self)
        # update the self.filelist.
        self.filelist.exclude_pattern(
            RENDERED_TMPD_PREFIX + ".*", is_regex=True
        )
        # but since mfname is already written we have to update it also.
        mfname = os.path.join(self.egg_info, "SOURCES.txt")
        if os.path.exists(mfname):
            with open(mfname) as fp:
                files = [
                    f for f in fp if not f.startswith(RENDERED_TMPD_PREFIX)
                ]
            with open(mfname, "w") as fp:
                fp.write("".join(files))


# TODO: Is there a better way to do this??
class InitsysInstallData(install):
    init_system = None
    user_options = install.user_options + [
        # This will magically show up in member variable 'init_sys'
        (
            "init-system=",
            None,
            "init system(s) to configure (%s) [default: None]"
            % ", ".join(INITSYS_TYPES),
        ),
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.init_system = ""

    def finalize_options(self):
        install.finalize_options(self)

        if self.init_system and isinstance(self.init_system, str):
            self.init_system = self.init_system.split(",")

        if len(self.init_system) == 0 and not platform.system().endswith(
            "BSD"
        ):
            self.init_system = ["systemd"]

        bad = [f for f in self.init_system if f not in INITSYS_TYPES]
        if len(bad) != 0:
            raise DistutilsError("Invalid --init-system: %s" % ",".join(bad))

        for system in self.init_system:
            # add data files for anything that starts with '<system>.'
            datakeys = [
                k for k in INITSYS_ROOTS if k.partition(".")[0] == system
            ]
            for k in datakeys:
                files = INITSYS_FILES[k]()
                if not files:
                    continue
                self.distribution.data_files.append((INITSYS_ROOTS[k], files))
        # Force that command to reinitialize (with new file list)
        self.distribution.reinitialize_command("install_data", True)


if not in_virtualenv():
    USR = "/" + USR
    ETC = "/" + ETC
    USR_LIB_EXEC = "/" + USR_LIB_EXEC
    LIB = "/" + LIB
    for k in INITSYS_ROOTS.keys():
        INITSYS_ROOTS[k] = "/" + INITSYS_ROOTS[k]

data_files = [
    (ETC + "/cloud", [render_tmpl("config/cloud.cfg.tmpl", is_yaml=True)]),
    (ETC + "/cloud/clean.d", glob("config/clean.d/*")),
    (ETC + "/cloud/cloud.cfg.d", glob("config/cloud.cfg.d/*")),
    (ETC + "/cloud/templates", glob("templates/*")),
    (
        USR_LIB_EXEC + "/cloud-init",
        [
            "tools/ds-identify",
            "tools/hook-hotplug",
            "tools/uncloud-init",
            "tools/write-ssh-key-fingerprints",
        ],
    ),
    (
        USR + "/share/bash-completion/completions",
        ["bash_completion/cloud-init"],
    ),
    (USR + "/share/doc/cloud-init", [f for f in glob("doc/*") if is_f(f)]),
    (
        USR + "/share/doc/cloud-init/examples",
        [f for f in glob("doc/examples/*") if is_f(f)],
    ),
    (
        USR + "/share/doc/cloud-init/examples/seed",
        [f for f in glob("doc/examples/seed/*") if is_f(f)],
    ),
]
if not platform.system().endswith("BSD"):
    RULES_PATH = pkg_config_read("udev", "udevdir")
    if not in_virtualenv():
        RULES_PATH = "/" + RULES_PATH

    data_files.extend(
        [
            (RULES_PATH + "/rules.d", [f for f in glob("udev/*.rules")]),
            (
                ETC + "/systemd/system/sshd-keygen@.service.d/",
                ["systemd/disable-sshd-keygen-if-cloud-init-active.conf"],
            ),
        ]
    )
# Use a subclass for install that handles
# adding on the right init system configuration files
cmdclass = {
    "install": InitsysInstallData,
    "egg_info": MyEggInfo,
}

requirements = read_requires()

setuptools.setup(
    name="cloud-init",
    version=get_version(),
    description="Cloud instance initialisation magic",
    author="Scott Moser",
    author_email="scott.moser@canonical.com",
    url="http://launchpad.net/cloud-init/",
    package_data={
        "": ["*.json"],
    },
    packages=setuptools.find_packages(exclude=["tests.*", "tests"]),
    scripts=["tools/cloud-init-per"],
    license="Dual-licensed under GPLv3 or Apache 2.0",
    data_files=data_files,
    install_requires=requirements,
    cmdclass=cmdclass,
    entry_points={
        "console_scripts": [
            "cloud-init = cloudinit.cmd.main:main",
            "cloud-id = cloudinit.cmd.cloud_id:main",
        ],
    },
)
